"""Tests for ResidualEngine, build_loss, audit, visualize, constitutive/scaling closures."""

import pytest
import jax
import jax.numpy as jnp
from moju.monitor import (
    AuditSpec,
    MonitorConfig,
    ResidualEngine,
    admissibility_level,
    audit,
    build_loss,
    list_constitutive_models,
    list_pi_constant_group_names,
    list_scaling_closure_ids,
    visualize,
)
from moju.monitor.auditor import DEFAULT_VISUALIZE_TITLE_TEST, DEFAULT_VISUALIZE_TITLE_TRAINING
from moju.monitor.closure_registry import (
    apply_closure_discrepancy_normalize,
    GROUP_FNS,
    MODEL_FNS,
    compute_implied_delta,
)
from moju.piratio.models import Models


class TestAdmissibilityLevel:
    def test_four_levels(self):
        assert admissibility_level(0.0) == "Non-Admissible"
        assert admissibility_level(0.39) == "Non-Admissible"
        assert admissibility_level(0.40) == "Low Admissibility"
        assert admissibility_level(0.69) == "Low Admissibility"
        assert admissibility_level(0.70) == "Moderate Admissibility"
        assert admissibility_level(0.89) == "Moderate Admissibility"
        assert admissibility_level(0.90) == "High Admissibility"
        assert admissibility_level(1.0) == "High Admissibility"

    def test_non_finite_unknown(self):
        assert admissibility_level(float("nan")) == "Unknown"
        assert admissibility_level(float("inf")) == "Unknown"


class TestNanTolerantAuditMetrics:
    def test_compute_log_step_metrics_category_zero_if_any_key_nan(self):
        import math

        from moju.monitor.auditor import _compute_log_step_metrics

        log = [
            {
                "index": 0,
                "rms": {
                    "constitutive/a/implied_delta": 1.0,
                    "constitutive/b/implied_delta": float("nan"),
                },
                "scale": {
                    "constitutive/a/implied_delta": 1.0,
                    "constitutive/b/implied_delta": 1.0,
                },
            }
        ]
        m = _compute_log_step_metrics(log)
        assert m[0]["category_admissibility_score"]["constitutive"] == 0.0
        assert "laws" not in m[0]["category_admissibility_score"]
        assert m[0]["overall_admissibility_score"] == 0.0

    def test_compute_log_step_metrics_all_nan_category_scores_zero_overall_zero(self):
        from moju.monitor.auditor import _compute_log_step_metrics

        log = [
            {
                "index": 0,
                "rms": {"constitutive/x/implied_delta": float("nan")},
                "scale": {"constitutive/x/implied_delta": 1.0},
            }
        ]
        m = _compute_log_step_metrics(log)
        assert m[0]["category_admissibility_score"]["constitutive"] == 0.0
        assert m[0]["overall_admissibility_score"] == 0.0

        rep = audit(log)
        assert rep["per_category"]["constitutive"] == 0.0
        assert rep["overall_admissibility_score"] == 0.0
        assert rep["overall_admissibility_level"] == "Non-Admissible"

    def test_rms_scalar_uses_nanmean(self):
        import math

        from moju.monitor.auditor import _rms_scalar

        x = jnp.array([1.0, jnp.nan, 3.0])
        r = float(_rms_scalar(x))
        assert math.isclose(r, math.sqrt(5.0), rel_tol=1e-5)


class TestResidualEngineResidualDict:
    def test_laws_only_when_no_audits(self, rtol, atol):
        core = ResidualEngine(
            constants={"L": 0.1},
            laws=[{"name": "mass_incompressible", "state_map": {"u_grad": "u_grad"}}],
        )
        state_pred = {"u_grad": jnp.array([[0.0, 1.0], [-1.0, 0.0]])}
        residuals = core.compute_residuals(state_pred)
        assert "laws" in residuals
        assert jnp.allclose(residuals["laws"]["mass_incompressible"], 0.0, rtol=rtol, atol=atol)
        assert "constitutive" not in residuals
        assert "scaling" not in residuals
        assert "data" not in residuals

    def test_log_appends_rms_per_key(self):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        state_pred = {"phi_laplacian": jnp.array(1.0)}
        core.compute_residuals(state_pred)
        core.compute_residuals(state_pred)
        assert len(core.log) == 2
        assert "laws/laplace_equation" in core.log[0]["rms"]

    def test_coord_snapshot_logged_from_state_coords(self):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        x = jnp.linspace(0.0, 1.0, 5)
        state_pred = {"phi_laplacian": jnp.array(1.0), "x": x}
        core.compute_residuals(state_pred)
        snap = core.log[-1].get("coord_snapshot") or {}
        assert "x" in snap
        assert len(snap["x"]) == 5
        assert all(isinstance(v, float) for v in snap["x"])

    def test_last_residuals_set_after_compute(self):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        state_pred = {"phi_laplacian": jnp.array(1.0)}
        r = core.compute_residuals(state_pred)
        assert core.last_residuals is r
        core.clear_log()
        assert core.last_residuals is None

    def test_clear_log_resets_entries_and_index(self):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        state_pred = {"phi_laplacian": jnp.array(1.0)}
        core.compute_residuals(state_pred)
        core.compute_residuals(state_pred)
        assert len(core.log) == 2
        core.clear_log()
        assert len(core.log) == 0
        core.compute_residuals(state_pred)
        assert len(core.log) == 1
        assert core.log[0].get("index") == 0

    def test_constitutive_sutherland_ref_delta_nonzero(self, rtol, atol):
        core = ResidualEngine(
            constants={"mu0": 1.8e-5, "T0": 273.0, "S": 110.4},
            laws=[],
            constitutive_audit=[
                {
                    "name": "sutherland_mu",
                    "output_key": "mu",
                    "state_map": {"T": "T", "mu0": "mu0", "T0": "T0", "S": "S"},
                }
            ],
        )
        T_pred = 300.0
        T_ref = 280.0
        mu_pred = 1.8e-5 * (T_pred / 273) ** 1.5 * (273 + 110.4) / (T_pred + 110.4)
        mu_ref = 1.8e-5 * (T_ref / 273) ** 1.5 * (273 + 110.4) / (T_ref + 110.4)
        state_pred = {"mu": mu_pred, "T": T_pred}
        state_ref = {"mu": mu_ref, "T": T_ref}
        residuals = core.compute_residuals(state_pred, state_ref=state_ref)
        assert "constitutive" in residuals
        assert "sutherland_mu/ref_delta" in residuals["constitutive"]
        assert abs(float(residuals["constitutive"]["sutherland_mu/ref_delta"])) > 0.0

    def test_scaling_pe_ref_delta_zero(self, rtol, atol):
        core = ResidualEngine(
            laws=[],
            scaling_audit=[
                {
                    "name": "pe",
                    "output_key": "Pe",
                    "state_map": {"re": "Re", "pr": "Pr"},
                }
            ],
        )
        Re, Pr = 100.0, 0.7
        Pe = Re * Pr
        state_pred = {"Pe": Pe, "Re": Re, "Pr": Pr}
        residuals = core.compute_residuals(state_pred, state_ref=dict(state_pred))
        assert "scaling" in residuals
        assert jnp.allclose(residuals["scaling"]["pe/ref_delta"], 0.0, rtol=rtol, atol=atol)

    def test_state_ref_adds_data_residual(self, rtol, atol):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        state_pred = {"phi_laplacian": jnp.array(0.5)}
        state_ref = {"phi_laplacian": jnp.array(0.0)}
        residuals = core.compute_residuals(state_pred, state_ref=state_ref)
        assert "data" in residuals
        assert jnp.allclose(residuals["data"]["phi_laplacian"], -0.5, rtol=rtol, atol=atol)

class TestBuildLoss:
    def test_cascaded_loss_scalar(self, rtol, atol):
        residual_dict = {
            "laws": {
                "mass_incompressible": jnp.array(0.0),
                "laplace_equation": jnp.array(0.0),
            }
        }
        loss = build_loss(residual_dict)
        assert jnp.allclose(loss, 0.0, rtol=rtol, atol=atol)

    def test_cascaded_loss_nonzero(self, rtol, atol):
        residual_dict = {"laws": {"laplace_equation": jnp.array(3.0)}}
        loss = build_loss(residual_dict)
        assert jnp.allclose(loss, 3.0, rtol=rtol, atol=atol)

    def test_build_loss_law_weights(self, rtol, atol):
        residual_dict = {"laws": {"a": jnp.array(1.0), "b": jnp.array(0.0)}}
        loss = build_loss(residual_dict, law_weights={"a": 1.0, "b": 0.0})
        assert jnp.allclose(loss, 1.0, rtol=rtol, atol=atol)

    def test_build_loss_differentiable(self):
        def loss_fn(phi_laplacian):
            return build_loss({"laws": {"laplace_equation": phi_laplacian}})

        grad = jax.grad(loss_fn)(jnp.array(2.0))
        assert grad is not None


class TestBuildLossBatch:
    def test_batch_law_residuals(self, rtol, atol):
        residual_dict = {"laws": {"mass_incompressible": jnp.zeros((5,))}}
        loss = build_loss(residual_dict)
        assert jnp.allclose(loss, 0.0, rtol=rtol, atol=atol)


class TestAudit:
    def test_audit_writes_back_to_log(self):
        log = [
            {"index": 0, "rms": {"laws/a": 2.0, "constitutive/x": 1.0}},
            {"index": 1, "rms": {"laws/a": 1.0, "constitutive/x": 0.5}},
        ]
        report = audit(log)
        assert "per_key" in report
        assert "r_norm" in log[0]

    def test_audit_r_ref_from_first_entry(self):
        log = [{"index": 0, "rms": {"k": 10.0}}, {"index": 1, "rms": {"k": 5.0}}]
        audit(log)
        assert log[1]["r_norm"]["k"] == 0.5

    def test_audit_uses_r_ref_as_scale_when_supplied(self):
        log = [{"index": 0, "rms": {"k": 100.0}}, {"index": 1, "rms": {"k": 4.0}}]
        audit(log, r_ref={"k": 8.0})
        assert log[1]["r_norm"]["k"] == 0.5
        assert log[0]["r_norm"]["k"] == 100.0 / 8.0

    def test_audit_uses_entry_scale_when_present(self):
        log = [
            {"index": 0, "rms": {"laws/a": 2.0}, "scale": {"laws/a": 4.0}},
            {"index": 1, "rms": {"laws/a": 1.0}, "scale": {"laws/a": 4.0}},
        ]
        audit(log)
        assert log[1]["r_norm"]["laws/a"] == 0.25
        assert log[0]["r_norm"]["laws/a"] == 0.5

    def test_engine_log_has_scale_and_audit_uses_it(self, rtol, atol):
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_xx"}}],
        )
        state_pred = {"phi_xx": jnp.array(1.0)}
        core.compute_residuals(state_pred)
        assert "scale" in core.log[-1]
        assert "laws/laplace_equation" in core.log[-1]["scale"]
        report = audit(core.log)
        r_norm = core.log[-1]["r_norm"]["laws/laplace_equation"]
        scale_k = core.log[-1]["scale"]["laws/laplace_equation"]
        rms = core.log[-1]["rms"]["laws/laplace_equation"]
        assert abs(r_norm - rms / scale_k) < 1e-6

    def test_default_scale_is_unit_for_laws_and_implied_delta(self, rtol, atol):
        """ND law and implied_delta keys use scale_k ≈ 1, not RMS(state primitives)."""
        P, R, T = jnp.array(1e6), jnp.array(287.0), jnp.array(300.0)
        rho = Models.ideal_gas_rho(P, R, T)
        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_xx"}}],
            constitutive_audit=[
                {
                    "name": "ideal_gas_rho",
                    "output_key": "rho",
                    "state_map": {"P": "P", "R": "R", "T": "T"},
                    "implied_value_key": "rho_implied",
                }
            ],
        )
        state_pred = {
            "phi_xx": jnp.array(2.0),
            "P": P,
            "R": R,
            "T": T,
            "rho": rho,
            "rho_implied": rho,
        }
        core.compute_residuals(state_pred)
        sc = core.log[-1]["scale"]
        assert abs(sc["laws/laplace_equation"] - 1.0) < 1e-9
        assert abs(sc["constitutive/ideal_gas_rho/implied_delta"] - 1.0) < 1e-9

    def test_constitutive_implied_scale_is_unity(self, rtol, atol):
        P, R = jnp.array(101325.0), jnp.array(287.0)
        T = jnp.array(290.0)
        rho = Models.ideal_gas_rho(P, R, T)
        core = ResidualEngine(
            laws=[],
            constitutive_audit=[
                {
                    "name": "ideal_gas_rho",
                    "output_key": "rho",
                    "state_map": {"P": "P", "R": "R", "T": "T"},
                    "implied_value_key": "rho_implied",
                }
            ],
        )
        state_pred = {"P": P, "R": R, "T": T, "rho": rho, "rho_implied": rho}
        core.compute_residuals(state_pred)
        sk = core.log[-1]["scale"]["constitutive/ideal_gas_rho/implied_delta"]
        assert abs(sk - 1.0) < 1e-6

    def test_audit_export_dir_pdf_with_new_categories(self, tmp_path):
        pytest.importorskip("reportlab")
        log = [
            {"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/implied_delta": 0.5, "scaling/pe/ref_delta": 0.1}},
            {"index": 1, "rms": {"laws/a": 0.5, "constitutive/m/implied_delta": 0.25, "scaling/pe/ref_delta": 0.05}},
        ]
        report = audit(log, export_dir=str(tmp_path))
        assert "per_key" in report
        dirs = [d for d in tmp_path.iterdir() if d.is_dir() and d.name.startswith("audit_")]
        assert len(dirs) == 1
        assert (dirs[0] / "report.pdf").exists()


class TestVisualize:
    def test_visualize_empty_log_returns_none(self):
        assert visualize([], backend="plotly") is None

    def test_visualize_backend_none_returns_none(self):
        assert visualize([{"index": 0, "rms": {"k": 1.0}}], backend="none") is None

    def test_visualize_matplotlib_backend_raises(self):
        with pytest.raises(ValueError, match="matplotlib"):
            visualize([{"index": 0, "rms": {"k": 1.0}, "scale": {}}], backend="matplotlib")

    def test_visualize_unknown_backend_raises(self):
        with pytest.raises(ValueError, match="Unknown visualize backend"):
            visualize([{"index": 0, "rms": {"k": 1.0}, "scale": {}}], backend="not_a_backend")

    def test_visualize_multi_panel_figure(self):
        pytest.importorskip("plotly")
        log = [
            {
                "index": 0,
                "rms": {
                    "laws/a": 1.0,
                    "constitutive/m/implied_delta": 0.5,
                    "scaling/pe/ref_delta": 0.1,
                    "data/T": 0.2,
                },
                "scale": {},
                "omitted": ["demo omit"],
                "inferred": [],
            },
            {
                "index": 1,
                "rms": {
                    "laws/a": 0.5,
                    "constitutive/m/implied_delta": 0.25,
                    "scaling/pe/ref_delta": 0.05,
                    "data/T": 0.1,
                },
                "scale": {},
            },
        ]
        fig = visualize(log, mode="training")
        assert fig is not None
        assert hasattr(fig, "data")
        assert len(fig.data) >= 7

    def test_visualize_r_norm_scale_linear(self):
        pytest.importorskip("plotly")
        log = [
            {
                "index": 0,
                "rms": {
                    "laws/a": 1.0,
                    "constitutive/m/implied_delta": 0.5,
                    "scaling/pe/ref_delta": 0.1,
                },
                "scale": {},
            },
            {
                "index": 1,
                "rms": {
                    "laws/a": 0.5,
                    "constitutive/m/implied_delta": 0.25,
                    "scaling/pe/ref_delta": 0.05,
                },
                "scale": {},
            },
        ]
        fig_p = visualize(log, mode="training", r_norm_scale="linear")
        assert fig_p is not None

    def test_build_monitor_visualize_bundle_and_studio_plotly_cards(self):
        pytest.importorskip("plotly")
        import numpy as np

        from moju.monitor.auditor import build_monitor_visualize_bundle
        from moju.monitor.visualize_plotly import (
            MOJU_STUDIO_DASHBOARD_CARD_HEIGHT,
            build_plotly_category_admissibility_bar_figure,
            build_plotly_law_rnorm_final_bar_figure,
            build_plotly_spatial_rnorm_heatmap_card,
        )

        log = [
            {
                "index": 0,
                "rms": {
                    "laws/a": 1.0,
                    "constitutive/m/implied_delta": 0.5,
                    "scaling/pe/ref_delta": 0.1,
                },
                "scale": {},
            },
            {
                "index": 1,
                "rms": {
                    "laws/a": 0.5,
                    "constitutive/m/implied_delta": 0.25,
                    "scaling/pe/ref_delta": 0.05,
                },
                "scale": {},
            },
        ]
        x = np.linspace(0, 1, 5)
        spatial_law = {"x": x, "values": {"laws/a": np.ones(5) * 0.2}}
        spatial_c = {"x": x, "values": {"constitutive/m/implied_delta": np.ones(5) * 0.1}}
        bundle = build_monitor_visualize_bundle(
            log,
            r_ref=None,
            max_legend_keys=16,
            spatial_law_panel=spatial_law,
            spatial_rnorm_panel=spatial_c,
            mode="training",
        )
        assert bundle is not None
        f1 = build_plotly_law_rnorm_final_bar_figure(bundle)
        f2 = build_plotly_category_admissibility_bar_figure(bundle)
        assert f1 is not None and f2 is not None
        assert f1.layout.height == MOJU_STUDIO_DASHBOARD_CARD_HEIGHT
        assert f2.layout.height == MOJU_STUDIO_DASHBOARD_CARD_HEIGHT
        law_bar = next((t for t in f1.data if getattr(t, "type", None) == "bar"), None)
        assert law_bar is not None
        marker_color = getattr(getattr(law_bar, "marker", None), "color", None)
        if isinstance(marker_color, (list, tuple)):
            assert all(c == "#7e57c2" for c in marker_color)
        else:
            assert marker_color == "#7e57c2"
        f3 = build_plotly_spatial_rnorm_heatmap_card(bundle["spatial"], colorscale="Jet")
        f4 = build_plotly_spatial_rnorm_heatmap_card(bundle["spatial_rnorm"], colorscale="Jet")
        assert f3 is not None and f4 is not None
        assert f3.layout.height == MOJU_STUDIO_DASHBOARD_CARD_HEIGHT
        assert f4.layout.height == MOJU_STUDIO_DASHBOARD_CARD_HEIGHT

    def test_build_monitor_visualize_bundle_uses_residuals_when_panels_none(self):
        import numpy as np

        from moju.monitor.auditor import build_monitor_visualize_bundle

        log = [
            {
                "index": 0,
                "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5},
                "scale": {"laws/a": 1.0, "constitutive/m/c": 1.0},
                "coord_snapshot": {"x": [0.0, 0.5, 1.0]},
            },
        ]
        residuals = {"laws": {"a": np.ones(3) * 0.1}, "constitutive": {"m/c": np.ones(3) * 0.05}}
        bundle = build_monitor_visualize_bundle(
            log,
            spatial_law_panel=None,
            spatial_rnorm_panel=None,
            mode="training",
            residuals=residuals,
            state_pred={},
        )
        assert bundle is not None
        assert bundle.get("spatial") is not None
        assert bundle.get("spatial_rnorm") is not None

    def test_maybe_build_spatial_panels_fills_missing_constitutive_side(self):
        import numpy as np

        from moju.monitor.auditor import _maybe_build_spatial_panels

        x = np.linspace(0, 1, 4)
        explicit_law = {"x": x, "values": {"laws/a": np.ones(4) * 0.2}}
        log = [
            {
                "index": 0,
                "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5},
                "scale": {"laws/a": 1.0, "constitutive/m/c": 1.0},
            },
        ]
        residuals = {"constitutive": {"m/c": np.ones(4) * 0.1}}
        law_p, rn_p = _maybe_build_spatial_panels(
            log,
            explicit_law,
            None,
            residuals,
            {"x": x},
            None,
            "x",
            True,
        )
        assert law_p is not None
        assert rn_p is not None
        assert "values" in rn_p and any(k.startswith("constitutive/") for k in (rn_p.get("values") or {}))

    def test_visualize_test_mode_uses_last_log_entry(self):
        pytest.importorskip("plotly")
        log = [
            {"index": 0, "rms": {"laws/a": 10.0}, "scale": {"laws/a": 1.0}},
            {"index": 1, "rms": {"laws/a": 0.1}, "scale": {"laws/a": 1.0}},
        ]
        fig = visualize(log, mode="test")
        assert fig is not None
        assert hasattr(fig, "data")

    def test_visualize_spatial_law_panel_plotly(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [
            {"index": 0, "rms": {"laws/a": 1.0}, "scale": {"laws/a": 1.0}},
            {"index": 1, "rms": {"laws/a": 0.5}, "scale": {"laws/a": 1.0}},
        ]
        x = np.linspace(0, 1, 5)
        fig = visualize(
            log,
            mode="training",
            spatial_law_panel={"x": x, "values": {"a": np.ones(5) * 0.2}},
        )
        assert fig is not None
        assert len(fig.data) >= 7
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 1

    def test_visualize_category_adm_high_scores_plotly(self):
        pytest.importorskip("plotly")
        log = [
            {
                "index": 0,
                "rms": {"laws/a": 0.0196, "constitutive/b": 0.0198},
                "scale": {"laws/a": 1.0, "constitutive/b": 1.0},
            },
            {
                "index": 1,
                "rms": {"laws/a": 0.0196, "constitutive/b": 0.0198},
                "scale": {"laws/a": 1.0, "constitutive/b": 1.0},
            },
        ]
        fig = visualize(log, mode="training")
        assert fig is not None
        assert len(fig.data) >= 5

    def test_build_visualize_bundle_category_training(self):
        from moju.monitor.auditor import _build_visualize_bundle

        log = [
            {"index": 0, "rms": {"laws/x": 1.0, "constitutive/a/b": 0.5, "scaling/y/z": 0.2}, "scale": {}},
            {"index": 1, "rms": {"laws/x": 0.5, "constitutive/a/b": 0.25, "scaling/y/z": 0.1}, "scale": {}},
        ]
        b = _build_visualize_bundle(log, None, None, 8, spatial_parsed=None, mode="training")
        assert b is not None
        ct = b["category_training"]
        assert "laws/x" in ct["laws"]["keys"]
        assert "constitutive/a/b" in ct["constitutive"]["keys"]
        assert "scaling" not in ct

    def test_visualize_invalid_mode_raises(self):
        with pytest.raises(ValueError, match="mode"):
            visualize([{"index": 0, "rms": {"k": 1.0}, "scale": {}}], mode="invalid")

    def test_visualize_plotly_returns_figure(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [
            {
                "index": 0,
                "rms": {
                    "laws/a": 1.0,
                    "constitutive/m/implied_delta": 0.5,
                    "scaling/pe/ref_delta": 0.1,
                    "data/T": 0.2,
                },
                "scale": {},
                "omitted": ["demo omit"],
                "inferred": [],
            },
            {
                "index": 1,
                "rms": {
                    "laws/a": 0.5,
                    "constitutive/m/implied_delta": 0.25,
                    "scaling/pe/ref_delta": 0.05,
                    "data/T": 0.1,
                },
                "scale": {},
            },
        ]
        x = np.linspace(0, 1, 5)
        spatial_law = {"x": x, "values": {"laws/a": np.ones(5) * 0.2}}
        spatial_c = {"x": x, "values": {"constitutive/m/implied_delta": np.ones(5) * 0.1}}
        fig = visualize(
            log,
            backend="plotly",
            mode="training",
            spatial_law_panel=spatial_law,
            spatial_rnorm_panel=spatial_c,
        )
        assert fig is not None
        assert hasattr(fig, "data")
        assert len(fig.data) >= 7
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 2

    def test_visualize_autofill_spatial_from_residuals(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [
            {"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5}, "scale": {"laws/a": 1.0, "constitutive/m/c": 1.0}},
        ]
        x = np.linspace(0, 1, 5)
        residuals = {"laws": {"a": np.ones(5) * 0.1}, "constitutive": {"m/c": np.ones(5) * 0.05}}
        pred = {"x": x}
        fig = visualize(
            log,
            mode="training",
            residuals=residuals,
            state_pred=pred,
        )
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 2

    def test_visualize_uses_engine_without_explicit_residuals(self):
        pytest.importorskip("plotly")

        core = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_laplacian"}}],
        )
        x = jnp.linspace(0.0, 1.0, 5)
        state_pred = {"phi_laplacian": jnp.ones(5) * 0.1, "x": x}
        core.compute_residuals(state_pred)
        log = list(core.log)
        assert log[-1].get("coord_snapshot")
        fig = visualize(log, mode="training", engine=core)
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 1

    def test_visualize_uses_coord_snapshot_when_state_pred_empty(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [
            {
                "index": 0,
                "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5},
                "scale": {"laws/a": 1.0, "constitutive/m/c": 1.0},
                "coord_snapshot": {"x": [0.0, 0.25, 0.5, 0.75, 1.0]},
            },
        ]
        residuals = {"laws": {"a": np.ones(5) * 0.1}, "constitutive": {"m/c": np.ones(5) * 0.05}}
        fig = visualize(
            log,
            mode="training",
            residuals=residuals,
            state_pred={},
        )
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 2

    def test_visualize_plotly_default_titles(self):
        pytest.importorskip("plotly")
        log = [
            {
                "index": 0,
                "rms": {
                    "laws/a": 1.0,
                    "constitutive/m/implied_delta": 0.5,
                    "scaling/pe/ref_delta": 0.1,
                },
                "scale": {},
            },
            {
                "index": 1,
                "rms": {
                    "laws/a": 0.5,
                    "constitutive/m/implied_delta": 0.25,
                    "scaling/pe/ref_delta": 0.05,
                },
                "scale": {},
            },
        ]
        fig_tr = visualize(log, backend="plotly", mode="training")
        assert DEFAULT_VISUALIZE_TITLE_TRAINING in (fig_tr.layout.title.text or "")
        fig_te = visualize(log, backend="plotly", mode="test")
        assert DEFAULT_VISUALIZE_TITLE_TEST in (fig_te.layout.title.text or "")

    def test_visualize_plotly_test_mode_spatial_row_placeholders_without_coords(self):
        pytest.importorskip("plotly")
        log = [{"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5}, "scale": {}}]
        fig = visualize(log, backend="plotly", mode="test")
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) == 0
        sc = [t for t in fig.data if getattr(t, "type", None) == "scatter"]
        assert any("spatial" in str(getattr(t, "text", "") or "").lower() for t in sc)

    def test_visualize_plotly_test_mode_heatmaps_with_residuals_only(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [{"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5}, "scale": {}}]
        residuals = {"laws": {"a": np.ones(5) * 0.1}, "constitutive": {"m/c": np.ones(5) * 0.05}}
        fig = visualize(log, backend="plotly", mode="test", residuals=residuals)
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) >= 2

    def test_visualize_plotly_test_mode_dual_spatial_heatmaps(self):
        pytest.importorskip("plotly")
        import numpy as np

        log = [{"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5}, "scale": {}}]
        x = np.linspace(0, 1, 5)
        spatial_law = {"x": x, "values": {"laws/a": np.ones(5) * 0.2}}
        spatial_c = {"x": x, "values": {"constitutive/m/c": np.ones(5) * 0.1}}
        fig = visualize(
            log,
            backend="plotly",
            mode="test",
            spatial_law_panel=spatial_law,
            spatial_rnorm_panel=spatial_c,
        )
        assert fig is not None
        hm = [t for t in fig.data if getattr(t, "type", None) == "heatmap"]
        assert len(hm) == 2

    def test_parse_spatial_law_panel_1d_2d_3d(self):
        import numpy as np

        from moju.monitor.auditor import _parse_spatial_law_panel

        x = np.linspace(0, 1, 3)
        y = np.linspace(0, 1, 4)
        z = np.linspace(0, 1, 5)
        o1 = _parse_spatial_law_panel({"x": x, "values": {"a": np.ones(3)}})
        assert o1 is not None and o1["kind"] == "1d"
        assert o1["Z"].shape == (1, 3)
        o2 = _parse_spatial_law_panel({"x": x, "y": y, "values": {"laplace": np.ones((4, 3))}})
        assert o2 is not None and o2["kind"] == "2d"
        assert o2["Z"].shape == (1, 4, 3)
        o3 = _parse_spatial_law_panel({"x": x, "y": y, "z": z, "values": {"laplace": np.ones((3, 4, 5))}})
        assert o3 is not None and o3["kind"] == "3d"
        assert o3["V"].shape == (1, 3, 4, 5)
        assert _parse_spatial_law_panel({"x": x, "y": y, "values": {"a": np.ones(3)}}) is None
        o2i = _parse_spatial_law_panel(
            {"x": x, "y": y, "values": {"a": np.ones((4, 3))}, "log_step_index": 7}
        )
        assert o2i is not None and o2i.get("log_step_index") == 7

    def test_build_plotly_spatial_card_2d_and_volume_3d(self):
        pytest.importorskip("plotly")
        import numpy as np

        from moju.monitor.auditor import _parse_spatial_law_panel
        from moju.monitor.visualize_plotly import build_plotly_spatial_rnorm_heatmap_card

        x, y = np.linspace(0, 1, 4), np.linspace(0, 1, 5)
        sp2 = _parse_spatial_law_panel({"x": x, "y": y, "values": {"laws/a": np.ones((5, 4)) * 0.2}})
        f2 = build_plotly_spatial_rnorm_heatmap_card(sp2)
        assert f2 is not None and f2.data
        assert getattr(f2.data[0], "type", None) == "heatmap"
        z = np.linspace(0, 1, 3)
        sp3 = _parse_spatial_law_panel(
            {
                "x": np.linspace(0, 1, 3),
                "y": np.linspace(0, 1, 4),
                "z": z,
                "values": {"laws/a": np.ones((3, 4, 3)) * 0.15},
            }
        )
        f3 = build_plotly_spatial_rnorm_heatmap_card(sp3, card_title="Vol")
        assert f3 is not None
        assert any(getattr(t, "type", None) == "volume" for t in f3.data)

    def test_build_plotly_spatial_card_1d_moves_key_to_subtitle_and_updates_colorbar(self):
        pytest.importorskip("plotly")
        import numpy as np

        from moju.monitor.auditor import _parse_spatial_law_panel
        from moju.monitor.visualize_plotly import build_plotly_spatial_rnorm_heatmap_card

        x = np.linspace(0, 1, 5)
        sp1 = _parse_spatial_law_panel({"x": x, "values": {"fourier_conduction": np.ones(5) * 0.2}})
        fig = build_plotly_spatial_rnorm_heatmap_card(sp1, card_title="Spatial |residual|")
        assert fig is not None

        assert fig.layout.yaxis.showticklabels is False

        title_text = fig.layout.title.text or ""
        assert "Fourier Conduction" in title_text

        hm = next(t for t in fig.data if getattr(t, "type", None) == "heatmap")
        cb_title_obj = getattr(getattr(hm, "colorbar", None), "title", None)
        cb_title_text = getattr(cb_title_obj, "text", None)
        assert cb_title_text == "log10(|residual| + ε)"

    def test_visualize_plotly_figure_title_override(self):
        pytest.importorskip("plotly")
        log = [
            {"index": 0, "rms": {"laws/a": 1.0, "constitutive/m/c": 0.5}, "scale": {}},
            {"index": 1, "rms": {"laws/a": 0.5, "constitutive/m/c": 0.25}, "scale": {}},
        ]
        fig = visualize(log, backend="plotly", mode="training", figure_title="Custom dashboard title")
        assert "Custom dashboard title" in (fig.layout.title.text or "")

    def test_visualize_training_all_residuals_bar_and_category_colors(self):
        pytest.importorskip("plotly")
        log = [
            {
                "index": 0,
                "rms": {
                    "laws/fourier_conduction": 1.0,
                    "constitutive/thermal_diffusivity/implied_delta": 0.5,
                },
                "scale": {},
            },
            {
                "index": 1,
                "rms": {
                    "laws/fourier_conduction": 0.5,
                    "constitutive/thermal_diffusivity/implied_delta": 0.25,
                },
                "scale": {},
            },
        ]
        fig = visualize(log, backend="plotly", mode="training")
        assert fig is not None

        # Overall admissibility axis stays percentage with two decimals.
        yaxes = [getattr(fig.layout, k) for k in dir(fig.layout) if k.startswith("yaxis")]
        assert any(
            getattr(getattr(ax, "title", None), "text", None) == "Admissibility (%)"
            and getattr(ax, "tickformat", None) == ".2f%"
            for ax in yaxes
        )

        # Vertical training residual bar includes constitutive implied residual too.
        vbars = [t for t in fig.data if getattr(t, "type", None) == "bar" and getattr(t, "orientation", None) != "h"]
        assert vbars
        x_labels = [str(x) for x in (vbars[0].x or [])]
        assert any("Fourier Conduction" in x for x in x_labels)
        assert any("Thermal diffusivity (implied)" in x for x in x_labels)

        xaxes = [getattr(fig.layout, k) for k in dir(fig.layout) if k.startswith("xaxis")]
        assert any(getattr(getattr(ax, "title", None), "text", None) == "All Residuals" for ax in xaxes)

        # Color mapping: laws=purple, constitutive=teal on non-overall residual traces.
        color_by_name = {}
        for tr in fig.data:
            if getattr(tr, "type", None) == "scatter" and getattr(tr, "mode", None) == "lines":
                nm = str(getattr(tr, "name", "") or "")
                line = getattr(tr, "line", None)
                color = getattr(line, "color", None)
                if nm:
                    color_by_name[nm] = color
        assert color_by_name.get("Fourier Conduction") == "#7e57c2"
        assert color_by_name.get("Thermal diffusivity (implied)") == "#009688"

    def test_compute_log_step_metrics_includes_data_category(self):
        log = [
            {
                "index": 0,
                "rms": {"laws/a": 1.0, "data/T": 0.5},
                "scale": {"laws/a": 1.0, "data/T": 1.0},
            }
        ]
        from moju.monitor.auditor import _compute_log_step_metrics

        m = _compute_log_step_metrics(log)
        assert "data" in m[0]["category_admissibility_score"]
        assert "laws" in m[0]["category_admissibility_score"]


class TestResidualEngineStateBuilder:
    def test_groups_enrich_state(self, rtol, atol):
        core = ResidualEngine(
            constants={"L": 0.1},
            groups=[{"name": "re", "state_map": {"u": "u", "L": "L", "rho": "rho", "mu": "mu"}, "output_key": "re"}],
        )
        state_pred = {"u": 1.0, "rho": 1000.0, "mu": 1.0}
        state = core._state_builder(state_pred)
        assert "re" in state
        assert jnp.allclose(state["re"], 100.0, rtol=rtol, atol=atol)


class TestCustomFn:
    def test_custom_law_fn(self, rtol, atol):
        def my_residual(x):
            return x - 1.0

        core = ResidualEngine(
            laws=[{"name": "my_law", "state_map": {"x": "x"}, "fn": my_residual}],
        )
        state_pred = {"x": jnp.array(2.0)}
        residuals = core.compute_residuals(state_pred)
        assert jnp.allclose(residuals["laws"]["my_law"], 1.0, rtol=rtol, atol=atol)
        assert jnp.allclose(build_loss(residuals), 1.0, rtol=rtol, atol=atol)

    def test_constitutive_custom_closure(self, rtol, atol):
        core = ResidualEngine(
            laws=[],
            constitutive_custom=[{"name": "my_c", "fn": lambda s, c: s["a"] - 2.0 * s["b"]}],
        )
        state_pred = {"a": jnp.array(4.0), "b": jnp.array(2.0)}
        residuals = core.compute_residuals(state_pred)
        assert jnp.allclose(residuals["constitutive"]["custom/my_c"], 0.0, rtol=rtol, atol=atol)

    def test_custom_group_fn_in_state(self, rtol, atol):
        def my_group(a, b):
            return a * b

        core = ResidualEngine(
            groups=[{"name": "my_ab", "state_map": {"a": "a", "b": "b"}, "output_key": "ab", "fn": my_group}],
        )
        state = core._state_builder({"a": jnp.array(3.0), "b": jnp.array(4.0)})
        assert jnp.allclose(state["ab"], 12.0, rtol=rtol, atol=atol)


class TestClosureDiscrepancyNormalize:
    def test_symmetric_zero_when_pred_matches_implied(self, rtol, atol):
        pred = jnp.array(3.0)
        implied = jnp.array(3.0)
        r = apply_closure_discrepancy_normalize(pred - implied, pred, implied)
        assert jnp.allclose(r, 0.0, rtol=rtol, atol=atol)

    def test_symmetric_scale_doubling_invariant(self, rtol, atol):
        pred = jnp.array(4.0)
        implied = jnp.array(2.0)
        r1 = apply_closure_discrepancy_normalize(pred - implied, pred, implied)
        r2 = apply_closure_discrepancy_normalize(
            2 * (pred - implied), 2 * pred, 2 * implied
        )
        assert jnp.allclose(r1, r2, rtol=rtol, atol=atol)

    def test_ref_scale_uses_ref_denominator(self, rtol, atol):
        pred = jnp.array(10.0)
        implied = jnp.array(6.0)
        ref = jnp.array(2.0)
        r = apply_closure_discrepancy_normalize(pred - implied, pred, implied, ref=ref)
        assert jnp.allclose(r, 4.0 / (1e-30 + 2.0), rtol=rtol, atol=atol)


class TestImpliedDeltaClosure:
    def test_compute_implied_delta_value_key(self, rtol, atol):
        fn, arg_names = MODEL_FNS["ideal_gas_rho"]
        P, R, T = jnp.array(1e5), jnp.array(287.0), jnp.array(300.0)
        rho_m = Models.ideal_gas_rho(P, R, T)
        merged = {"P": P, "R": R, "T": T, "rho_alt": rho_m}
        r = compute_implied_delta(
            fn=fn,
            arg_names=arg_names,
            state_map={"P": "P", "R": "R", "T": "T"},
            state_pred=merged,
            constants={},
            implied_value_key="rho_alt",
        )
        assert r is not None
        assert jnp.allclose(r, 0.0, rtol=rtol, atol=atol)

    def test_compute_implied_delta_missing_key_returns_none(self):
        fn, arg_names = MODEL_FNS["ideal_gas_rho"]
        r = compute_implied_delta(
            fn=fn,
            arg_names=arg_names,
            state_map={"P": "P", "R": "R", "T": "T"},
            state_pred={"P": 1.0, "R": 1.0, "T": 1.0},
            constants={},
            implied_value_key="rho_missing",
        )
        assert r is None

    def test_engine_implied_delta_ideal_gas(self, rtol, atol):
        P, R = jnp.array(101325.0), jnp.array(287.0)
        T = jnp.array(290.0)
        rho = Models.ideal_gas_rho(P, R, T)
        core = ResidualEngine(
            laws=[],
            constitutive_audit=[
                {
                    "name": "ideal_gas_rho",
                    "output_key": "rho",
                    "state_map": {"P": "P", "R": "R", "T": "T"},
                    "implied_value_key": "rho_implied",
                }
            ],
        )
        res = core.compute_residuals(
            {"P": P, "R": R, "T": T, "rho": rho, "rho_implied": rho}
        )
        assert "constitutive" in res
        assert jnp.allclose(
            res["constitutive"]["ideal_gas_rho/implied_delta"], 0.0, rtol=rtol, atol=atol
        )

    def test_ref_delta_without_predicted_spatial(self, rtol, atol):
        """ref_delta runs when state_ref is set even if predicted_* are empty."""
        core = ResidualEngine(
            laws=[],
            constitutive_audit=[
                {
                    "name": "ideal_gas_rho",
                    "output_key": "rho",
                    "state_map": {"P": "P", "R": "R", "T": "T"},
                }
            ],
        )
        P, R, T = jnp.array(1e5), jnp.array(287.0), jnp.array(300.0)
        rho1 = Models.ideal_gas_rho(P, R, T)
        rho2 = Models.ideal_gas_rho(P * 1.01, R, T)
        state_pred = {"P": P, "R": R, "T": T, "rho": rho1}
        state_ref = {"P": P * 1.01, "R": R, "T": T, "rho": rho2}
        res = core.compute_residuals(state_pred, state_ref=state_ref)
        assert "ideal_gas_rho/ref_delta" in res["constitutive"]

    def test_implied_both_key_and_fn_raises(self):
        with pytest.raises(ValueError, match="only one of implied"):
            ResidualEngine(
                laws=[],
                constitutive_audit=[
                    {
                        "name": "ideal_gas_rho",
                        "output_key": "rho",
                        "state_map": {"P": "P", "R": "R", "T": "T"},
                        "implied_value_key": "x",
                        "implied_fn": lambda s, c: s.get("x"),
                    }
                ],
            )

    def test_law_linked_implied_contributes_to_audit_scoring(self):
        """Law-linked implied rows are included in constitutive category/overall scoring."""
        engine = ResidualEngine(
            laws=[
                {
                    "name": "fourier_conduction",
                    "state_map": {
                        "T_t": "T_t",
                        "T_laplacian": "T_laplacian",
                        "fo": "fo",
                        "t": "t",
                        "L": "L",
                    },
                }
            ],
            groups=[{"name": "fo", "output_key": "fo", "state_map": {"alpha": "alpha", "t": "t", "L": "L"}}],
            law_implied_audits=True,
        )
        state_pred = {
            "T_t": jnp.array([1.0]),
            "T_laplacian": jnp.array([1.0]),
            "alpha": jnp.array([1.0]),
            "t": jnp.array([1.0]),
            "L": jnp.array([1.0]),
            "k": jnp.array([1.0]),
            "rho": jnp.array([1.0]),
            "cp": jnp.array([1.0]),
        }
        _ = engine.compute_residuals(state_pred)
        rep = audit(engine.log)
        assert "constitutive" in rep["per_category"]
        assert rep["per_category"]["constitutive"] >= 0.0


class TestRegistryHelpers:
    def test_list_constitutive_models(self):
        names = list_constitutive_models()
        assert "sutherland_mu" in names
        assert "thermal_diffusivity" in names
        assert "smagorinsky_nu_t" in names
        assert "k_epsilon_nu_t" in names
        assert "k_omega_nu_t" in names

    def test_list_scaling_closure_ids(self):
        ids = list_scaling_closure_ids()
        assert "pe" in ids
        assert "fo" in ids


class TestMonitorConfig:
    def test_to_from_dict_roundtrip(self):
        cfg = MonitorConfig(
            constants={"cp": 1.0},
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_xx"}}],
            constitutive_audit=[
                AuditSpec(
                    name="sutherland_mu",
                    output_key="mu",
                    state_map={"T": "T", "mu0": "mu0", "T0": "T0", "S": "S"},
                )
            ],
        )
        d = cfg.to_dict()
        cfg2 = MonitorConfig.from_dict(d)
        assert cfg2.to_dict() == d


class TestRequiredKeys:
    def test_required_state_keys_union(self):
        engine = ResidualEngine(
            laws=[{"name": "laplace_equation", "state_map": {"phi_laplacian": "phi_xx"}}],
            scaling_audit=[
                {
                    "name": "pe",
                    "output_key": "Pe",
                    "state_map": {"re": "Re", "pr": "Pr"},
                }
            ],
        )
        state_keys = engine.required_state_keys()
        assert "phi_xx" in state_keys
        assert "Re" in state_keys and "Pr" in state_keys and "Pe" in state_keys

    def test_scaling_pe_ref_delta_with_matching_ref(self, rtol, atol):
        engine = ResidualEngine(
            laws=[],
            scaling_audit=[
                {
                    "name": "pe",
                    "output_key": "Pe",
                    "state_map": {"re": "u", "pr": "T"},
                }
            ],
        )
        state_pred = {"u": 1.0, "T": 2.0, "Pe": 2.0}
        residuals = engine.compute_residuals(state_pred, state_ref=dict(state_pred))
        assert "scaling" in residuals
        assert jnp.allclose(residuals["scaling"]["pe/ref_delta"], 0.0, rtol=rtol, atol=atol)


def _re_pi_spec(*, compare_keys=("out",), scale_c=10.0):
    return {
        "name": "re",
        "output_key": "Re",
        "state_map": {"u": "u", "L": "L", "rho": "rho", "mu": "mu"},
        "invariance_pi_constant": True,
        "invariance_compare_keys": list(compare_keys),
        "invariance_scale_c": scale_c,
    }


class TestPiConstantClosure:
    def test_list_pi_constant_group_names(self):
        names = list_pi_constant_group_names()
        assert "re" in names and "pr" in names and "pe" in names
        assert "st" in names and "gr" in names and "we" in names
        assert set(names) == set(GROUP_FNS.keys())

    def test_apply_pi_constant_recipe_re_preserves_re(self):
        from moju.monitor.pi_constant_recipes import (
            GROUP_PI_CONSTANT_RECIPES,
            apply_pi_constant_recipe,
        )

        const = {"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0}
        sm = {"u": "u", "L": "L", "rho": "rho", "mu": "mu"}
        out = apply_pi_constant_recipe(const, GROUP_PI_CONSTANT_RECIPES["re"], sm, 10.0)
        Re0 = const["rho"] * const["u"] * const["L"] / const["mu"]
        Re1 = float(out["rho"] * out["u"] * out["L"] / out["mu"])
        assert abs(Re0 - Re1) < 1e-9

    def test_apply_pi_constant_recipe_c_must_exceed_one(self):
        from moju.monitor.pi_constant_recipes import (
            GROUP_PI_CONSTANT_RECIPES,
            apply_pi_constant_recipe,
        )

        const = {"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0}
        sm = {"u": "u", "L": "L", "rho": "rho", "mu": "mu"}
        with pytest.raises(ValueError, match="c > 1"):
            apply_pi_constant_recipe(const, GROUP_PI_CONSTANT_RECIPES["re"], sm, 1.0)

    def test_engine_init_requires_recipe_or_compare_keys(self):
        with pytest.raises(ValueError, match="invariance_compare_keys"):
            ResidualEngine(
                constants={"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0},
                laws=[],
                groups=[],
                scaling_audit=[_re_pi_spec(compare_keys=())],
                state_builder=lambda m, p, col, ct: {"out": jnp.array(1.0)},
            )

    def test_all_registered_groups_have_pi_recipe(self):
        from moju.monitor.pi_constant_recipes import assert_pi_recipes_cover_all_groups

        assert_pi_recipes_cover_all_groups()

    def test_engine_init_invariance_c_must_exceed_one(self):
        with pytest.raises(ValueError, match="invariance_scale_c"):
            ResidualEngine(
                constants={"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0},
                laws=[],
                groups=[],
                scaling_audit=[_re_pi_spec(scale_c=1.0)],
                state_builder=lambda m, p, col, ct: {"out": ct["L"] / ct["mu"]},
            )

    def test_path_b_forbidden_when_pi_enabled(self):
        def sb(model, params, collocation, constants):
            return {"out": constants["L"] / constants["mu"]}

        engine = ResidualEngine(
            constants={"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0},
            laws=[],
            groups=[],
            scaling_audit=[_re_pi_spec()],
            state_builder=sb,
        )
        with pytest.raises(ValueError, match="Path A"):
            engine.compute_residuals({"out": jnp.array(0.5), "Re": jnp.array(0.25), "u": jnp.array(1.0)})

    def test_path_a_pi_residual_zero_when_invariant(self, rtol, atol):
        def sb(model, params, collocation, constants):
            return {"out": constants["L"] / constants["mu"]}

        engine = ResidualEngine(
            constants={"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0},
            laws=[],
            groups=[],
            scaling_audit=[_re_pi_spec()],
            state_builder=sb,
        )
        residuals = engine.compute_residuals(None, model=0, params=0, collocation={})
        r = residuals["scaling"]["re/pi_constant"]
        assert jnp.allclose(r, 0.0, rtol=rtol, atol=atol)

    def test_pi_constant_scale_uses_mean_abs_scaled_branch(self):
        def sb(model, params, collocation, constants):
            return {"out": constants["L"]}

        engine = ResidualEngine(
            constants={"L": 1.0, "mu": 2.0, "rho": 1.0, "u": 1.0},
            laws=[],
            groups=[],
            scaling_audit=[_re_pi_spec()],
            state_builder=sb,
        )
        engine.compute_residuals(None, model=0, params=0, collocation={})
        entry = engine.log[-1]
        scale = entry["scale"]["scaling/re/pi_constant"]
        assert scale > 5.0
        rms = entry["rms"]["scaling/re/pi_constant"]
        assert rms > 0.01


class TestAuditSpecPiFieldsRoundtrip:
    def test_monitor_config_scaling_audit_pi_fields(self):
        spec = AuditSpec(
            name="re",
            output_key="Re",
            state_map={"u": "u", "L": "L", "rho": "rho", "mu": "mu"},
            invariance_pi_constant=True,
            invariance_compare_keys=["out"],
            invariance_scale_c=7.0,
        )
        cfg = MonitorConfig(constants={}, scaling_audit=[spec])
        d = cfg.to_dict()
        cfg2 = MonitorConfig.from_dict(d)
        s2 = cfg2.scaling_audit[0]
        assert s2.invariance_pi_constant is True
        assert s2.invariance_compare_keys == ["out"]
        assert s2.invariance_scale_c == 7.0


class TestKwargsFromStateLawHint:
    def test_missing_derived_key_includes_fd_hint(self):
        from moju.monitor.auditor import _kwargs_from_state

        with pytest.raises(KeyError, match="fill_law_fd"):
            _kwargs_from_state(
                {},
                {},
                {"T_laplacian": "T_laplacian"},
                law_context="fourier_conduction",
            )

    def test_missing_non_derived_key_omits_fd_sentence(self):
        from moju.monitor.auditor import _kwargs_from_state

        with pytest.raises(KeyError) as ei:
            _kwargs_from_state(
                {},
                {},
                {"Fo": "Fo"},
                law_context="fourier_conduction",
            )
        assert "fill_law_fd" not in str(ei.value)

    def test_placeholder_string_for_derived_key_treated_as_missing(self):
        from moju.monitor.auditor import _kwargs_from_state

        with pytest.raises(KeyError, match="fill_law_fd"):
            _kwargs_from_state(
                {},
                {"T_laplacian": "T_laplacian"},
                {"T_laplacian": "T_laplacian"},
                law_context="fourier_conduction",
            )

    def test_non_placeholder_string_raises_clear_type_error(self):
        from moju.monitor.auditor import _kwargs_from_state

        with pytest.raises(TypeError, match="resolved to string value"):
            _kwargs_from_state(
                {},
                {"T_laplacian": "oops"},
                {"T_laplacian": "T_laplacian"},
                law_context="fourier_conduction",
            )



"""
Law-linked implied constitutive / scaling audits.

For selected :class:`Laws.*` entries, Moju can auto-append ``constitutive_audit`` /
``scaling_audit`` rows whose ``implied_fn`` recomputes a quantity by rearranging the
law using ``state_pred`` (and the law's ``state_map``). Residuals use the existing
``implied_delta`` closure: ``F(catalog args) - implied``.

We avoid pairing **both** a group and a constitutive closure that are algebraically
equivalent given fixed scales (e.g. Fourier: only ``thermal_diffusivity``, not a
separate ``fo`` implied row).

See README "Law-linked implied audits" and :func:`merge_law_implied_audit_specs`.
"""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import jax.numpy as jnp

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _val(
    state: Dict[str, Any],
    constants: Dict[str, Any],
    law_sm: Dict[str, str],
    arg: str,
) -> Any:
    key = law_sm.get(arg)
    if key is None:
        return None
    v = state.get(key)
    if v is None:
        v = constants.get(key)
    return v


def _state_or_const(state: Dict[str, Any], constants: Dict[str, Any], key: str) -> Any:
    v = state.get(key)
    if v is None:
        v = constants.get(key)
    return v


def _safe_ratio(num: Any, den: Any, *, eps: float = 1e-20) -> jnp.ndarray:
    num = jnp.asarray(num, dtype=jnp.result_type(jnp.asarray(num), jnp.float32))
    den = jnp.asarray(den, dtype=jnp.result_type(jnp.asarray(den), jnp.float32))
    return jnp.where(jnp.abs(den) > eps, num / den, jnp.nan)


def _vec_norm_last(x: Any) -> jnp.ndarray:
    a = jnp.asarray(x)
    if a.ndim == 0:
        return jnp.abs(a)
    return jnp.sqrt(jnp.sum(a**2, axis=-1))


# ---------------------------------------------------------------------------
# Implied quantity makers: (law_state_map) -> implied_fn(state, constants)
# ---------------------------------------------------------------------------


def implied_alpha_fourier_conduction(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """α_implied = T_t / T_laplacian from :func:`Laws.fourier_conduction` rearrangement."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        T_t = _val(state, constants, law_sm, "T_t")
        T_lap = _val(state, constants, law_sm, "T_laplacian")
        if T_t is None or T_lap is None:
            return None
        return _safe_ratio(T_t, T_lap)

    return implied_fn


def implied_fo_mass_fick_diffusion(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Fo_m implied = φ_t * t / (φ_laplacian * L²) from Fick's law."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        pt = _val(state, constants, law_sm, "phi_t")
        pl = _val(state, constants, law_sm, "phi_laplacian")
        t = _val(state, constants, law_sm, "t")
        L = _val(state, constants, law_sm, "L")
        if pt is None or pl is None or t is None or L is None:
            return None
        return _safe_ratio(pt * t, pl * (L**2))

    return implied_fn


def implied_D_fick_diffusion(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """D_implied = phi_t / phi_laplacian from :func:`Laws.fick_diffusion` rearrangement."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        pt = _val(state, constants, law_sm, "phi_t")
        pl = _val(state, constants, law_sm, "phi_laplacian")
        if pt is None or pl is None:
            return None
        return _safe_ratio(pt, pl)

    return implied_fn


def implied_st_wave_wave_equation(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """St_wave_implied = ω L / c with c² = φ_tt / φ_laplacian (wave equation rearrangement)."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        ptt = _val(state, constants, law_sm, "phi_tt")
        pl = _val(state, constants, law_sm, "phi_laplacian")
        omega = _val(state, constants, law_sm, "omega")
        L = _val(state, constants, law_sm, "L")
        if ptt is None or pl is None or omega is None or L is None:
            return None
        c_sq = _safe_ratio(ptt, pl)
        c = jnp.sqrt(jnp.maximum(c_sq, 0.0))
        return jnp.asarray(omega) * jnp.asarray(L) / (c + 1e-30)

    return implied_fn


def implied_c_wave_equation(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """c_implied = sqrt(phi_tt / phi_laplacian) from :func:`Laws.wave_equation`."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        ptt = _val(state, constants, law_sm, "phi_tt")
        pl = _val(state, constants, law_sm, "phi_laplacian")
        if ptt is None or pl is None:
            return None
        c_sq = _safe_ratio(ptt, pl)
        return jnp.sqrt(jnp.maximum(c_sq, 0.0))

    return implied_fn


def implied_pe_advection_diffusion(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Pe_implied ≈ |∇²φ| / |φ_t + u·∇φ| from scalar transport rearrangement."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        phi_t = _val(state, constants, law_sm, "phi_t")
        u = _val(state, constants, law_sm, "u")
        g = _val(state, constants, law_sm, "phi_grad")
        lap = _val(state, constants, law_sm, "phi_laplacian")
        if phi_t is None or u is None or g is None or lap is None:
            return None
        u = jnp.asarray(u)
        g = jnp.asarray(g)
        if u.shape[-1] != g.shape[-1]:
            return None
        adv = jnp.sum(g * u, axis=-1)
        rhs = jnp.asarray(phi_t) + adv
        return _vec_norm_last(lap) / (_vec_norm_last(rhs) + 1e-30)

    return implied_fn


def implied_kappa_advection_diffusion(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """kappa_implied = U*L*(phi_t + u·grad(phi))/phi_laplacian (effective diffusivity)."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        phi_t = _val(state, constants, law_sm, "phi_t")
        u = _val(state, constants, law_sm, "u")
        g = _val(state, constants, law_sm, "phi_grad")
        lap = _val(state, constants, law_sm, "phi_laplacian")
        L = _state_or_const(state, constants, "L")
        if phi_t is None or u is None or g is None or lap is None or L is None:
            return None
        u = jnp.asarray(u)
        g = jnp.asarray(g)
        if u.shape[-1] != g.shape[-1]:
            return None
        adv = jnp.sum(g * u, axis=-1)
        rhs = jnp.asarray(phi_t) + adv
        U = _vec_norm_last(u)
        return jnp.asarray(U) * jnp.asarray(L) * _safe_ratio(rhs, lap)

    return implied_fn


def implied_re_momentum_navier_stokes(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Re_implied ≈ ‖∇²u‖ / ‖u_t + u·∇u + ∇p‖ (diagnostic when the NS residual is small)."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        u_t = _val(state, constants, law_sm, "u_t")
        u = _val(state, constants, law_sm, "u")
        u_grad = _val(state, constants, law_sm, "u_grad")
        p_grad = _val(state, constants, law_sm, "p_grad")
        u_lap = _val(state, constants, law_sm, "u_laplacian")
        if u_t is None or u is None or u_grad is None or p_grad is None or u_lap is None:
            return None
        u = jnp.asarray(u)
        adv = jnp.einsum("...ij,...j->...i", jnp.asarray(u_grad), u)
        rhs = jnp.asarray(u_t) + adv + jnp.asarray(p_grad)
        return _vec_norm_last(u_lap) / (_vec_norm_last(rhs) + 1e-30)

    return implied_fn


def implied_mu_from_re_balance(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """
    μ_implied = ρ u_char L / Re_implied with u_char = ‖u‖, using the same Re balance as
    :func:`implied_re_momentum_navier_stokes`. Compare to ``mu`` in state (e.g. from Sutherland).
    """

    re_fn = implied_re_momentum_navier_stokes(law_sm)

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        re_i = re_fn(state, constants)
        if re_i is None:
            return None
        u = _val(state, constants, law_sm, "u")
        rho = state.get("rho")
        if rho is None:
            rho = constants.get("rho")
        L = state.get("L")
        if L is None:
            L = constants.get("L")
        if u is None or rho is None or L is None:
            return None
        u = jnp.asarray(u)
        u_char = _vec_norm_last(u)
        rho = jnp.asarray(rho)
        L = jnp.asarray(L)
        return rho * u_char * L / (jnp.asarray(re_i) + 1e-30)

    return implied_fn


def implied_mu_from_re_fn(
    law_sm: Dict[str, str],
    re_fn: Callable[[Dict[str, Any], Dict[str, Any]], Optional[jnp.ndarray]],
) -> Callable[..., Any]:
    """mu_implied = rho * |u| * L / Re_implied from a supplied implied-Re function."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        re_i = re_fn(state, constants)
        if re_i is None:
            return None
        u = _val(state, constants, law_sm, "u")
        if u is None:
            u = _state_or_const(state, constants, "u")
        rho = _state_or_const(state, constants, "rho")
        L = _state_or_const(state, constants, "L")
        if u is None or rho is None or L is None:
            return None
        u_char = _vec_norm_last(u)
        return jnp.asarray(rho) * jnp.asarray(u_char) * jnp.asarray(L) / (jnp.asarray(re_i) + 1e-30)

    return implied_fn


def implied_mu_momentum_navier_stokes(law_sm: Dict[str, str]) -> Callable[..., Any]:
    return implied_mu_from_re_fn(law_sm, implied_re_momentum_navier_stokes(law_sm))


def implied_re_stokes_flow(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Re_implied ≈ ‖∇²u‖ / ‖∇p‖ for Stokes."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        p_grad = _val(state, constants, law_sm, "p_grad")
        u_lap = _val(state, constants, law_sm, "u_laplacian")
        if p_grad is None or u_lap is None:
            return None
        return _vec_norm_last(u_lap) / (_vec_norm_last(p_grad) + 1e-30)

    return implied_fn


def implied_mu_stokes_flow(law_sm: Dict[str, str]) -> Callable[..., Any]:
    return implied_mu_from_re_fn(law_sm, implied_re_stokes_flow(law_sm))


def implied_re_burgers_equation(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Re_implied ≈ (U L) ‖∇²u‖ / ‖u_t + (u·∇)u‖."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        u_t = _val(state, constants, law_sm, "u_t")
        u = _val(state, constants, law_sm, "u")
        u_grad = _val(state, constants, law_sm, "u_grad")
        u_lap = _val(state, constants, law_sm, "u_laplacian")
        U = _val(state, constants, law_sm, "U")
        L = _val(state, constants, law_sm, "L")
        if u_t is None or u is None or u_grad is None or u_lap is None or U is None or L is None:
            return None
        u = jnp.asarray(u)
        adv = jnp.einsum("...ij,...j->...i", jnp.asarray(u_grad), u)
        rhs = jnp.asarray(u_t) + adv
        return (jnp.asarray(U) * jnp.asarray(L)) * (
            _vec_norm_last(u_lap) / (_vec_norm_last(rhs) + 1e-30)
        )

    return implied_fn


def implied_mu_burgers_equation(law_sm: Dict[str, str]) -> Callable[..., Any]:
    return implied_mu_from_re_fn(law_sm, implied_re_burgers_equation(law_sm))


def implied_viscous_acceleration_incompressible(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Implied Newtonian viscous acceleration u_t + (u·∇)u + ∇p from the momentum balance."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        u_t = _val(state, constants, law_sm, "u_t")
        u = _val(state, constants, law_sm, "u")
        u_grad = _val(state, constants, law_sm, "u_grad")
        p_grad = _val(state, constants, law_sm, "p_grad")
        if u_t is None or u is None or u_grad is None or p_grad is None:
            return None
        adv = jnp.einsum("...ij,...j->...i", jnp.asarray(u_grad), jnp.asarray(u))
        return jnp.asarray(u_t) + adv + jnp.asarray(p_grad)

    return implied_fn


def implied_viscous_acceleration_compressible(law_sm: Dict[str, str]) -> Callable[..., Any]:
    """Implied ρ(u_t + u·∇u) + ∇p for the simplified compressible momentum form."""

    def implied_fn(state: Dict[str, Any], constants: Dict[str, Any]) -> Optional[jnp.ndarray]:
        rho = _val(state, constants, law_sm, "rho")
        u_t = _val(state, constants, law_sm, "u_t")
        u = _val(state, constants, law_sm, "u")
        u_grad = _val(state, constants, law_sm, "u_grad")
        p_grad = _val(state, constants, law_sm, "p_grad")
        if rho is None or u_t is None or u is None or u_grad is None or p_grad is None:
            return None
        adv = jnp.einsum("...ij,...j->...i", jnp.asarray(u_grad), jnp.asarray(u))
        rho_b = jnp.asarray(rho)[..., jnp.newaxis]
        return rho_b * (jnp.asarray(u_t) + adv) + jnp.asarray(p_grad)

    return implied_fn



# ---------------------------------------------------------------------------
# Registry: law_name -> list of audit spec dict fragments (merged into engine lists)
# ---------------------------------------------------------------------------

# ``residual_basename`` becomes ``{basename}/implied_delta`` under constitutive/ or scaling/.
# ``include_ref_delta``: when False, skip F(pred)-F(ref) even if state_ref is set.

_LAW_IMPLIED_ROWS: Dict[str, List[Dict[str, Any]]] = {
    "fourier_conduction": [
        {
            "category": "constitutive",
            "name": "thermal_diffusivity",
            "output_key": "alpha",
            "state_map": {"k": "k", "rho": "rho", "cp": "cp"},
            "implied_maker": implied_alpha_fourier_conduction,
            "residual_basename": "thermal_diffusivity/law_fourier_conduction",
            "include_ref_delta": True,
        },
    ],
    "fick_diffusion": [
        {
            "category": "constitutive",
            "name": "mass_diffusivity",
            "output_key": "D",
            "state_map": {"fo_mass": "fo_mass", "t": "t", "L": "L"},
            "implied_maker": implied_D_fick_diffusion,
            "residual_basename": "mass_diffusivity/law_fick_diffusion",
            "include_ref_delta": True,
        },
    ],
    "wave_equation": [
        {
            "category": "constitutive",
            "name": "wave_speed_from_st",
            "output_key": "c",
            "state_map": {"omega": "omega", "L": "L", "st_wave": "st_wave"},
            "implied_maker": implied_c_wave_equation,
            "residual_basename": "wave_speed_from_st/law_wave_equation",
            "include_ref_delta": True,
        },
    ],
    "advection_diffusion": [
        {
            "category": "constitutive",
            "name": "scalar_diffusivity_from_pe",
            "output_key": "kappa",
            "state_map": {"u": "u", "L": "L", "pe": "pe"},
            "implied_maker": implied_kappa_advection_diffusion,
            "residual_basename": "scalar_diffusivity_from_pe/law_advection_diffusion",
            "include_ref_delta": True,
        },
    ],
    "momentum_navier_stokes": [
        {
            "category": "constitutive",
            "name": "dynamic_viscosity_from_re",
            "output_key": "mu",
            "state_map": {"rho": "rho", "u": "u", "L": "L", "re": "re"},
            "implied_maker": implied_mu_momentum_navier_stokes,
            "residual_basename": "dynamic_viscosity_from_re/law_momentum_navier_stokes",
            "include_ref_delta": True,
        },
    ],
    "stokes_flow": [
        {
            "category": "constitutive",
            "name": "dynamic_viscosity_from_re",
            "output_key": "mu",
            "state_map": {"rho": "rho", "u": "u", "L": "L", "re": "re"},
            "implied_maker": implied_mu_stokes_flow,
            "residual_basename": "dynamic_viscosity_from_re/law_stokes_flow",
            "include_ref_delta": True,
        },
    ],
    "burgers_equation": [
        {
            "category": "constitutive",
            "name": "dynamic_viscosity_from_re",
            "output_key": "mu",
            "state_map": {"rho": "rho", "u": "u", "L": "L", "re": "re"},
            "implied_maker": implied_mu_burgers_equation,
            "residual_basename": "dynamic_viscosity_from_re/law_burgers_equation",
            "include_ref_delta": True,
        },
    ],
    "momentum_incompressible_newtonian_laplacian": [
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_k_omega",
            "output_key": "viscous_accel",
            "state_map": {
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "k": "k",
                "omega": "omega",
                "omega0": "omega0",
            },
            "implied_maker": implied_viscous_acceleration_incompressible,
            "residual_basename": "turbulent_viscous_acceleration_k_omega/law_momentum_incompressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_k_epsilon",
            "output_key": "viscous_accel",
            "state_map": {
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "C_mu": "C_mu",
                "k": "k",
                "epsilon": "epsilon",
                "eps0": "eps0",
            },
            "implied_maker": implied_viscous_acceleration_incompressible,
            "residual_basename": "turbulent_viscous_acceleration_k_epsilon/law_momentum_incompressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_smagorinsky",
            "output_key": "viscous_accel",
            "state_map": {
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "Cs": "Cs",
                "Delta": "Delta",
                "strain_rate_magnitude": "strain_rate_magnitude",
            },
            "implied_maker": implied_viscous_acceleration_incompressible,
            "residual_basename": "turbulent_viscous_acceleration_smagorinsky/law_momentum_incompressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
    ],
    "momentum_compressible_newtonian_laplacian": [
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_compressible_k_omega",
            "output_key": "viscous_accel",
            "state_map": {
                "rho": "rho",
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "k": "k",
                "omega": "omega",
                "omega0": "omega0",
            },
            "implied_maker": implied_viscous_acceleration_compressible,
            "residual_basename": "turbulent_viscous_acceleration_compressible_k_omega/law_momentum_compressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_compressible_k_epsilon",
            "output_key": "viscous_accel",
            "state_map": {
                "rho": "rho",
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "C_mu": "C_mu",
                "k": "k",
                "epsilon": "epsilon",
                "eps0": "eps0",
            },
            "implied_maker": implied_viscous_acceleration_compressible,
            "residual_basename": "turbulent_viscous_acceleration_compressible_k_epsilon/law_momentum_compressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
        {
            "category": "constitutive",
            "name": "turbulent_viscous_acceleration_compressible_smagorinsky",
            "output_key": "viscous_accel",
            "state_map": {
                "rho": "rho",
                "u_laplacian": "u_laplacian",
                "nu_molecular": "nu_molecular",
                "Cs": "Cs",
                "Delta": "Delta",
                "strain_rate_magnitude": "strain_rate_magnitude",
            },
            "implied_maker": implied_viscous_acceleration_compressible,
            "residual_basename": "turbulent_viscous_acceleration_compressible_smagorinsky/law_momentum_compressible_newtonian_laplacian",
            "include_ref_delta": True,
        },
    ],
}


# Best-effort policy: these laws currently have no law-linked implied row because a
# stable constitutive/scaling rearrangement is not encoded in the monitor registry.
_LAW_IMPLIED_UNSUPPORTED_REASONS: Dict[str, str] = {
    "laplace_equation": "no explicit constitutive/scaling term to rearrange",
    "poisson_equation": "requires source-term closure choice not represented by one catalog model",
    "helmholtz_equation": "depends on forcing/wavenumber framing rather than one implied closure target",
    "schrodinger_steady": "complex-valued amplitude/phase closure not encoded as one implied scalar",
    "laplace_beltrami": "metric/geometry operators require chart-specific closure treatment",
    "mass_incompressible": "constraint law (divergence-free) has no direct constitutive target",
    "mass_compressible": "continuity balance has no single implied constitutive quantity",
    "euler_momentum": "inviscid law has no viscosity constitutive closure target",
    "darcy_flow": "permeability/drag closure requires porous-medium model context",
    "brinkman_extension": "mixed porous/viscous closure needs medium-specific constitutive choices",
    "viscous_dissipation": "dissipation source does not map to one implied constitutive/scaling scalar",
    "hookes_law_residual": "elastic constitutive inversion is material-model specific",
}


def list_laws_with_implied_diagnostics() -> Tuple[str, ...]:
    """Law registry names that contribute auto implied audits."""
    return tuple(sorted(_LAW_IMPLIED_ROWS.keys()))

def law_implied_unsupported_reasons() -> Dict[str, str]:
    """
    Best-effort coverage map for laws without auto implied rows.

    Returns law-name -> short reason. This documents intentional gaps in the
    law-linked implied registry rather than leaving them implicit.
    """
    return dict(_LAW_IMPLIED_UNSUPPORTED_REASONS)


def effective_audit_specs_for_fragment(d: Dict[str, Any]) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Return ``(constitutive_audit, scaling_audit)`` as the engine sees them: law-linked rows
    (when ``law_implied_audits`` is true) merged with fragment lists (see
    :func:`merge_fragment_law_implied_audit_specs`). Used by Studio dependency planning
    and JSON fragments without ``implied_fn``.
    """
    lic, lis = merge_law_implied_audit_specs(
        d.get("laws") or [],
        enabled=bool(d.get("law_implied_audits", True)),
    )
    frag_c = list(d.get("constitutive_audit") or [])
    frag_s = list(d.get("scaling_audit") or [])
    mc, rc = merge_fragment_law_implied_audit_specs(lic, frag_c)
    ms, rs = merge_fragment_law_implied_audit_specs(lis, frag_s)
    return mc + rc, ms + rs


def merge_law_implied_audit_specs(
    laws_spec: Sequence[Dict[str, Any]],
    *,
    enabled: bool = True,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Build extra ``constitutive_audit`` and ``scaling_audit`` dict rows for
    :class:`ResidualEngine` from the selected laws.

    Rows include ``implied_fn`` (not JSON-serializable) and ``residual_basename`` for
    unique log keys. When ``enabled`` is False, returns ``([], [])``.
    """
    if not enabled:
        return [], []
    constitutive: List[Dict[str, Any]] = []
    scaling: List[Dict[str, Any]] = []
    seen_basenames: set[str] = set()

    for law in laws_spec:
        law_name = str(law.get("name") or "")
        rows = _LAW_IMPLIED_ROWS.get(law_name)
        if not rows:
            continue
        law_sm = dict(law.get("state_map") or {})
        for row in rows:
            basename = str(row["residual_basename"])
            if basename in seen_basenames:
                continue
            seen_basenames.add(basename)
            implied_maker = row["implied_maker"]
            d = {
                "name": row["name"],
                "output_key": row["output_key"],
                "state_map": dict(row["state_map"]),
                "implied_fn": implied_maker(law_sm),
                "residual_basename": basename,
                "include_ref_delta": bool(row.get("include_ref_delta", True)),
            }
            if row["category"] == "constitutive":
                constitutive.append(d)
            else:
                scaling.append(d)

    return constitutive, scaling


def merge_fragment_law_implied_audit_specs(
    prepended: List[Dict[str, Any]],
    from_config: List[Dict[str, Any]],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Return ``(prepended, extras)``. Rows in ``from_config`` that declare the same
    ``residual_basename`` as a prepended law-linked row are dropped (legacy duplicate
    suppression); all other config rows are kept in ``extras``.
    """
    prep_bn = {str(r["residual_basename"]) for r in prepended if r.get("residual_basename")}
    extras: List[Dict[str, Any]] = []
    for d in from_config:
        bn = d.get("residual_basename")
        if bn and str(bn) in prep_bn:
            continue
        extras.append(d)
    return list(prepended), extras



"""
π-constant scaling recipes for Groups.* scaling audits.

Each recipe lists rows ``(arg_name, kind[, power])`` where ``kind`` is ``multiply_c`` or
``divide_c``, and optional integer ``power`` (default 1) means scale by ``c**power`` or
``c**(-power)`` respectively. Combined effect per state key is ``v * c**total_exponent``.

Recipes are **canonical one-parameter orbits** (not unique). For ``gr``, gravity is
fixed inside ``Groups.gr`` (see module docstring in :mod:`moju.piratio.groups`).
"""

from __future__ import annotations

from typing import Any, Dict, List, Tuple, Union

import jax
import jax.numpy as jnp

# Row: (arg_name, multiply_c|divide_c) or (arg_name, multiply_c|divide_c, positive_int_power)
RecipeRow = Union[Tuple[str, str], Tuple[str, str, int]]

GROUP_PI_CONSTANT_RECIPES: Dict[str, List[RecipeRow]] = {
    # Re = rho*u*L/mu
    "re": [("L", "multiply_c"), ("mu", "multiply_c")],
    # Pr = cp*mu/k
    "pr": [("mu", "multiply_c"), ("k", "multiply_c")],
    # Nu = h*L/k  ->  h'*k' with L fixed, or scale h,k together
    "nu": [("h", "multiply_c"), ("k", "multiply_c")],
    # Gr = g*beta*dT*L^3/nu^2 (g fixed in Groups.gr). L^2 and nu^3 -> L^6/nu^6 cancels.
    "gr": [("L", "multiply_c", 2), ("nu", "multiply_c", 3)],
    # Ma = u/a
    "ma": [("u", "multiply_c"), ("a", "multiply_c")],
    # We = rho*u^2*L/sigma  ->  u*c, sigma*c^2
    "we": [("u", "multiply_c"), ("sigma", "multiply_c", 2)],
    # Pe = re*pr
    "pe": [("re", "multiply_c"), ("pr", "divide_c")],
    # St = f*L/u  ->  f*c, u*c
    "st": [("f", "multiply_c"), ("u", "multiply_c")],
    # Bi = h*L/k_solid
    "bi": [("h", "multiply_c"), ("k_solid", "multiply_c")],
    # Fo = alpha*t/L^2  ->  alpha*c, t*c, L*c
    "fo": [("alpha", "multiply_c"), ("t", "multiply_c"), ("L", "multiply_c")],
    # Sc = nu/D
    "sc": [("nu", "multiply_c"), ("D", "multiply_c")],
    # Le = sc/pr  ->  scale sc and pr by the same c
    "le": [("sc", "multiply_c"), ("pr", "multiply_c")],
    # Kn = lambda_mfp/L
    "kn": [("lambda_mfp", "multiply_c"), ("L", "multiply_c")],
    # Bo = rho*g*L^2/sigma  ->  L*c, sigma*c^2
    "bo": [("L", "multiply_c"), ("sigma", "multiply_c", 2)],
    # Ca = mu*u/sigma  ->  mu*c, sigma*c
    "ca": [("mu", "multiply_c"), ("sigma", "multiply_c")],
    # Eu = dp/(rho*u^2)  ->  dp*c^2, u*c
    "eu": [("dp", "multiply_c", 2), ("u", "multiply_c")],
    # Da = K/L^2  ->  K*c^2, L*c
    "da": [("K", "multiply_c", 2), ("L", "multiply_c")],
    # Ec = u^2/(cp*dT)
    "ec": [("u", "multiply_c"), ("cp", "multiply_c"), ("dT", "multiply_c")],
    # Fo_mass = D*t/L^2
    "fo_mass": [("D", "multiply_c"), ("t", "multiply_c"), ("L", "multiply_c")],
    # wavenumber = k*L  ->  k/c, L*c
    "wavenumber": [("k", "divide_c"), ("L", "multiply_c")],
    # Pe_mass = re*sc
    "pe_mass": [("re", "multiply_c"), ("sc", "divide_c")],
    # St_wave = omega*L/c_wave
    "st_wave": [("omega", "multiply_c"), ("c", "multiply_c")],
    # poisson_rhs_pi = L^2 * source / (epsilon * phi_ref)
    "poisson_rhs_pi": [("L", "multiply_c"), ("source", "divide_c", 2)],
    # schrodinger_kinetic_length_squared = 2*m*L^2/h_bar^2  ->  L*c, h_bar*c
    "schrodinger_kinetic_length_squared": [("L", "multiply_c"), ("h_bar", "multiply_c")],
}


def list_pi_constant_group_names() -> List[str]:
    return sorted(GROUP_PI_CONSTANT_RECIPES.keys())


def _row_to_exponent(row: Tuple[Any, ...]) -> Tuple[str, int]:
    if len(row) == 2:
        arg_name, kind = row
        n = 1
    elif len(row) == 3:
        arg_name, kind, n = row
        if not isinstance(n, int) or n < 1:
            raise ValueError(f"recipe row third element must be int >= 1, got {row!r}")
    else:
        raise ValueError(f"recipe row must be length 2 or 3, got {row!r}")
    if kind == "multiply_c":
        return arg_name, n
    if kind == "divide_c":
        return arg_name, -n
    raise ValueError(f"recipe kind must be 'multiply_c' or 'divide_c', got {kind!r}")


def apply_pi_constant_recipe(
    constants: Dict[str, Any],
    recipe: List[RecipeRow],
    state_map: Dict[str, str],
    c: Union[float, Any],
) -> Dict[str, Any]:
    """
    Copy ``constants`` and scale entries referenced by ``recipe`` via ``state_map``.

    Exponents for the same state key (from different args) add. Each state key must
    exist in ``constants`` (Path A: engine.constants).
    """
    c_arr = jnp.asarray(c)
    c_float = float(jax.device_get(c_arr).item())
    if c_float <= 1.0:
        raise ValueError(f"π-constant scaling requires c > 1, got {c!r}")
    c = c_arr

    exp_by_state_key: Dict[str, int] = {}
    for row in recipe:
        arg_name, exp = _row_to_exponent(row)
        if arg_name not in state_map:
            raise KeyError(f"π-constant: arg {arg_name!r} missing from state_map")
        sk = state_map[arg_name]
        exp_by_state_key[sk] = exp_by_state_key.get(sk, 0) + exp

    out = dict(constants)
    for sk, total_exp in exp_by_state_key.items():
        if sk not in out:
            raise KeyError(
                f"π-constant: state key {sk!r} not found in constants; "
                "put scaled parameters in ResidualEngine.constants for Path A."
            )
        v = jnp.asarray(out[sk])
        out[sk] = v * (c ** total_exp)
    return out


def assert_pi_recipes_cover_all_groups() -> None:
    """Dev check: registry keys match recipes (call from tests)."""
    from moju.monitor.closure_registry import GROUP_FNS

    reg = set(GROUP_FNS.keys())
    rec = set(GROUP_PI_CONSTANT_RECIPES.keys())
    if reg != rec:
        missing = sorted(reg - rec)
        extra = sorted(rec - reg)
        raise AssertionError(f"GROUP_FNS vs recipes mismatch missing={missing} extra={extra}")



"""Tests for moju.piratio.Laws (conservation-law residuals)."""

import pytest
import jax.numpy as jnp
from moju.piratio import Groups, Laws


class TestLawsMassIncompressible:
    """div(u) = 0 residual."""

    def test_zero_gradient_gives_zero_residual(self, rtol, atol):
        """Constant velocity field: u_grad = 0 => div u = 0."""
        u_grad = jnp.zeros((2, 2))
        residual = Laws.mass_incompressible(u_grad)
        assert jnp.allclose(residual, 0.0, rtol=rtol, atol=atol)

    def test_trace_zero_gives_zero_residual(self, rtol, atol):
        """Solenoidal field: trace(u_grad) = 0 => div u = 0."""
        # e.g. u = (y, -x) => u_grad = [[0,1],[-1,0]], trace = 0
        u_grad = jnp.array([[0.0, 1.0], [-1.0, 0.0]])
        residual = Laws.mass_incompressible(u_grad)
        assert jnp.allclose(residual, 0.0, rtol=rtol, atol=atol)

    def test_non_solenoidal_gives_nonzero_residual(self, rtol, atol):
        """trace != 0 => residual != 0."""
        u_grad = jnp.array([[1.0, 0.0], [0.0, 1.0]])  # trace = 2
        residual = Laws.mass_incompressible(u_grad)
        assert not jnp.allclose(residual, 0.0, atol=1e-6)
        assert jnp.allclose(residual, 2.0, rtol=rtol, atol=atol)

    def test_batch_mass_incompressible(self, rtol, atol):
        """Batch of velocity gradients returns 1D residual."""
        u_grad = jnp.zeros((3, 2, 2))  # 3 points, 2x2 gradient each
        residual = Laws.mass_incompressible(u_grad)
        assert residual.shape == (3,)
        assert jnp.allclose(residual, 0.0, rtol=rtol, atol=atol)


class TestLawsLaplaceAndWave:
    """Simple PDE residuals."""

    def test_laplace_equation_residual(self, rtol, atol):
        """Laplace equation: residual = Laplacian(phi). Zero when Laplacian is 0."""
        residual = Laws.laplace_equation(0.0)
        assert jnp.allclose(residual, 0.0, rtol=rtol, atol=atol)

    def test_wave_equation_residual_zero(self, rtol, atol):
        """Wave equation: phi_tt - c^2 Laplacian(phi) = 0 with c = omega*L/st_wave. Zero when both terms cancel."""
        phi_tt = 1.0
        phi_laplacian = 1.0
        omega = 1.0
        L = 1.0
        st_wave = 1.0  # c = omega*L/st_wave = 1.0
        residual = Laws.wave_equation(phi_tt, phi_laplacian, st_wave, omega, L)
        expected = phi_tt - 1.0 * phi_laplacian
        assert jnp.allclose(residual, expected, rtol=rtol, atol=atol)
        assert jnp.allclose(residual, 0.0, rtol=rtol, atol=atol)


class TestLawsFourierConduction:
    """Fourier conduction: T_t - alpha * T_laplacian with optional grid broadcasting."""

    def test_broadcasts_alpha_when_t_1d_and_fields_2d(self, rtol, atol):
        """Mesh time ``t(n_t,)`` with ``T_t``, ``T_laplacian`` (n_t, n_x) — alpha must broadcast."""
        nt, nx = 4, 5
        T_t = jnp.ones((nt, nx))
        T_laplacian = jnp.ones((nt, nx))
        t = jnp.linspace(0.5, 2.0, nt)
        fo = jnp.array(0.1)
        L = jnp.array(1.0)
        out = Laws.fourier_conduction(T_t, T_laplacian, fo, t, L)
        assert out.shape == (nt, nx)
        alpha = fo * (L**2) / t
        alpha2 = jnp.broadcast_to(
            alpha.reshape(nt, 1), (nt, nx)
        )
        assert jnp.allclose(out, T_t - alpha2 * T_laplacian, rtol=rtol, atol=atol)
        assert jnp.all(jnp.isfinite(out))


class TestLawsSchrodingerSteady:
    """Nondimensional steady Schrödinger matches scaled dimensional TISE."""

    def test_nd_residual_equals_k_times_dimensional_residual(self, rtol, atol):
        # Order-one m, L, h_bar so float32 JAX does not underflow h_bar**2.
        lap_dim = jnp.array(-1.5)
        V = jnp.array(2.0)
        E = jnp.array(1.0)
        psi = jnp.array(0.5 + 0.2j)
        m = jnp.array(1.0)
        h_bar = jnp.array(1.0)
        L = jnp.array(1.0)
        K = Groups.schrodinger_kinetic_length_squared(m, L, h_bar)
        lap_L2 = (L**2) * lap_dim
        r_new = Laws.schrodinger_steady(lap_L2, V, E, psi, K)
        r_dim = (-(h_bar**2) / (2 * m)) * lap_dim + (V - E) * psi
        assert jnp.allclose(r_new, K * r_dim, rtol=rtol, atol=atol)


class TestLawsStokesFlow:
    """Stokes flow residual: grad p - (1/Re) Laplacian(u) = 0."""

    def test_stokes_flow_residual_shape(self):
        """Stokes residual is vector of same dimension as p_grad."""
        p_grad = jnp.array([1.0, 0.0])
        u_laplacian = jnp.array([0.0, 0.0])
        re = 100.0
        residual = Laws.stokes_flow(p_grad, u_laplacian, re)
        assert residual.shape == (2,)


class TestLawsNewtonianLaplacianMomentum:
    def test_incompressible_balance_zero(self, rtol, atol):
        u_t = jnp.array([0.0, 0.0])
        u = jnp.array([1.0, 0.0])
        u_grad = jnp.zeros((2, 2))
        p_grad = jnp.array([0.0, 0.0])
        u_laplacian = jnp.array([0.0, 0.0])
        nu = jnp.array(0.1)
        r = Laws.momentum_incompressible_newtonian_laplacian(
            u_t, u, u_grad, p_grad, u_laplacian, nu
        )
        assert jnp.allclose(r, 0.0, rtol=rtol, atol=atol)

    def test_compressible_balance_zero(self, rtol, atol):
        rho = jnp.array(1.2)
        u_t = jnp.array([0.0, 0.0])
        u = jnp.array([0.0, 0.0])
        u_grad = jnp.zeros((2, 2))
        p_grad = jnp.array([0.0, 0.0])
        u_laplacian = jnp.array([0.0, 0.0])
        nu = jnp.array(0.0)
        r = Laws.momentum_compressible_newtonian_laplacian(
            rho, u_t, u, u_grad, p_grad, u_laplacian, nu
        )
        assert jnp.allclose(r, 0.0, rtol=rtol, atol=atol)



"""PDF export for Physics Admissibility Report. Requires reportlab (pip install moju[report])."""

from __future__ import annotations

import datetime
import json
from typing import Any, Dict, List, Optional, Tuple

from moju.monitor.visualize_labels import format_admissibility_pct, pretty_residual_key


try:
    from reportlab.lib import colors
    from reportlab.lib.pagesizes import letter
    from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
    from reportlab.lib.units import inch
    from reportlab.platypus import Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle
    _REPORTLAB_AVAILABLE = True
except ImportError:
    _REPORTLAB_AVAILABLE = False

# Section headers for first path segment of log keys (laws/..., constitutive/..., scaling/..., data/...)
_CATEGORY_HEADERS: Dict[str, str] = {
    "laws": "Governing Laws",
    "constitutive": "Constitutive relations",
    "scaling": "Scaling and similarity",
    "data": "Data",
    "groups": "Scaling and similarity (legacy key)",
    "models": "Constitutive (legacy key)",
}


def _group_keys_by_category(per_key: Dict[str, Any]) -> List[Tuple[str, List[Tuple[str, Any]]]]:
    buckets: Dict[str, List[Tuple[str, Any]]] = {
        "laws": [],
        "constitutive": [],
        "scaling": [],
        "data": [],
        "groups": [],
        "models": [],
    }
    for key, data in per_key.items():
        if "/" in key:
            prefix, _rest = key.split("/", 1)
            if prefix in buckets:
                buckets[prefix].append((pretty_residual_key(key), data))
    order = ("laws", "constitutive", "scaling", "data", "groups", "models")
    result: List[Tuple[str, List[Tuple[str, Any]]]] = []
    for cat in order:
        if buckets[cat]:
            result.append((_CATEGORY_HEADERS[cat], buckets[cat]))
    return result


def _residual_dict_to_json_serializable(residual_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Recursively convert residual dict (JAX/NumPy arrays) to JSON-serializable lists."""
    try:
        import jax
    except ImportError:
        jax = None
    import numpy as np

    out: Dict[str, Any] = {}
    for key, val in residual_dict.items():
        if isinstance(val, dict):
            out[key] = _residual_dict_to_json_serializable(val)
        elif hasattr(val, "tolist"):
            try:
                arr = val
                if jax is not None and hasattr(arr, "__jax_array__"):
                    arr = jax.device_get(arr)
                out[key] = np.asarray(arr).tolist()
            except Exception:
                out[key] = str(val)
        else:
            out[key] = val
    return out


def write_residuals_json(residual_dict: Dict[str, Any], path: str) -> None:
    """Write residual dict to a JSON file (arrays converted to lists)."""
    data = _residual_dict_to_json_serializable(residual_dict)
    with open(path, "w") as f:
        json.dump(data, f, indent=2)


_FOOTER_LEFT = "Moju is developed by Ifimo Lab at Ifimo Analytics"
_FOOTER_RIGHT = "This report is a heuristic and not a certification."


def _footer_canvas(canvas: Any, doc: Any) -> None:
    """Draw footer on each page: credit (left), disclaimer (right)."""
    canvas.saveState()
    canvas.setFont("Helvetica", 8)
    canvas.setFillColor(colors.HexColor("#555555"))
    y = 0.5 * inch
    canvas.drawString(doc.leftMargin, y, _FOOTER_LEFT)
    x_right = doc.pagesize[0] - doc.rightMargin
    canvas.drawRightString(x_right, y, _FOOTER_RIGHT)
    canvas.restoreState()


def write_audit_pdf(
    report: Dict[str, Any],
    path: str,
    model_name: Optional[str] = None,
    model_id: Optional[str] = None,
) -> None:
    """
    Write a Physics Admissibility Report PDF to the given path.

    :param report: Dict from audit() with per_key, overall_admissibility_score, overall_admissibility_level.
    :param path: Output file path (e.g. .pdf).
    :param model_name: Optional model name for the header.
    :param model_id: Optional model ID for the header.
    """
    if not _REPORTLAB_AVAILABLE:
        raise ImportError(
            "reportlab is required for PDF export. Install with: pip install moju[report] or pip install reportlab"
        )

    per_key = report.get("per_key", {})
    per_category = report.get("per_category", {})
    overall_score = report.get("overall_admissibility_score", 0.0)
    overall_level = report.get("overall_admissibility_level", "Non-Admissible")

    doc = SimpleDocTemplate(
        path,
        pagesize=letter,
        rightMargin=0.75 * inch,
        leftMargin=0.75 * inch,
        topMargin=0.75 * inch,
        bottomMargin=0.75 * inch,
    )
    styles = getSampleStyleSheet()
    title_style = ParagraphStyle(
        "ReportTitle",
        parent=styles["Heading1"],
        fontSize=18,
        spaceAfter=12,
    )
    heading_style = ParagraphStyle(
        "SectionHeading",
        parent=styles["Heading2"],
        fontSize=12,
        spaceBefore=14,
        spaceAfter=6,
    )
    body_style = styles["Normal"]

    story: List[Any] = []

    # Title
    story.append(Paragraph("Physics Admissibility Report", title_style))
    story.append(Spacer(1, 6))

    # Optional model name, ID, date
    meta_parts: List[str] = []
    if model_name:
        meta_parts.append(f"Model: {model_name}")
    if model_id:
        meta_parts.append(f"Model ID: {model_id}")
    meta_parts.append(f"Date: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M')}")
    if meta_parts:
        story.append(Paragraph(" &nbsp; &nbsp; ".join(meta_parts), body_style))
        story.append(Spacer(1, 12))

    story.append(
        Paragraph(
            "<i>These metrics are consistency indicators for declared laws, constitutive closures, "
            "and scaling identities on the evaluated sample. They are not a certification of "
            "physical correctness.</i>",
            body_style,
        )
    )
    story.append(Spacer(1, 12))

    # Overall score and level
    story.append(Paragraph("Overall", heading_style))
    score_text = f"Admissibility score: {format_admissibility_pct(float(overall_score))} &nbsp; — &nbsp; {overall_level}"
    story.append(Paragraph(score_text, body_style))
    story.append(Spacer(1, 16))

    if per_category:
        story.append(Paragraph("Category summary", heading_style))
        table_data = [["Category", "Score"]]
        for key, label in (("laws", "Governing laws"), ("constitutive", "Constitutive"), ("scaling", "Scaling/similarity")):
            if key in per_category:
                table_data.append([label, format_admissibility_pct(float(per_category[key]))])
        t = Table(table_data, colWidths=[3.0 * inch, 1.2 * inch])
        t.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.HexColor("#e0e0e0")),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("FONTSIZE", (0, 0), (-1, -1), 10),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 8),
                    ("TOPPADDING", (0, 0), (-1, 0), 8),
                    ("GRID", (0, 0), (-1, -1), 0.5, colors.grey),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)
        story.append(
            Paragraph(
                "<i>Category scores are geometric means of the per-metric admissibility scores in that section.</i>",
                body_style,
            )
        )
        story.append(Spacer(1, 16))

    # Per-key metrics by category
    grouped = _group_keys_by_category(per_key)
    for section_header, items in grouped:
        story.append(Paragraph(section_header, heading_style))
        table_data = [["Metric", "Score", "Level"]]
        for label, data in items:
            score = data.get("admissibility_score", 0.0)
            level = data.get("admissibility_level", "Non-Admissible")
            table_data.append([label, format_admissibility_pct(float(score)), level])
        t = Table(table_data, colWidths=[2.5 * inch, 1 * inch, 2.2 * inch])
        t.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.HexColor("#e0e0e0")),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("FONTSIZE", (0, 0), (-1, -1), 10),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 8),
                    ("TOPPADDING", (0, 0), (-1, 0), 8),
                    ("GRID", (0, 0), (-1, -1), 0.5, colors.grey),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)
        story.append(Spacer(1, 8))

    doc.build(
        story,
        onFirstPage=_footer_canvas,
        onLaterPages=_footer_canvas,
    )



"""
Predefined JAX-differentiable constitutive closure residuals.

Each closure returns pred_field - Models.*(...) (or balance form). Return None if required
keys are missing from state/constants (closure skipped).

State keys (examples):
  sutherland_mu: mu, T + constants mu0, T0, S
  thermal_diffusivity: alpha, k, rho, cp
  ideal_gas_rho: rho, P, R, T
"""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp

from moju.piratio.models import Models


def _val(state: Dict[str, Any], constants: Dict[str, Any], key: str) -> Any:
    v = state.get(key)
    if v is None:
        v = constants.get(key)
    return v


def _all_present(state: Dict[str, Any], constants: Dict[str, Any], keys: Tuple[str, ...]) -> bool:
    for k in keys:
        if _val(state, constants, k) is None:
            return False
    return True


def _closure_sutherland_direct(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("mu", "T", "mu0", "T0", "S")):
        return None
    mu = _val(state, constants, "mu")
    T = _val(state, constants, "T")
    mu0 = _val(state, constants, "mu0")
    T0 = _val(state, constants, "T0")
    S = _val(state, constants, "S")
    return mu - Models.sutherland_mu(T, mu0, T0, S)


def _closure_vft_direct(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("mu", "T", "A", "B", "T0_v")):
        return None
    return _val(state, constants, "mu") - Models.vft_mu(
        _val(state, constants, "T"),
        _val(state, constants, "A"),
        _val(state, constants, "B"),
        _val(state, constants, "T0_v"),
    )


def _closure_ideal_gas_rho(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("rho", "P", "R", "T")):
        return None
    return _val(state, constants, "rho") - Models.ideal_gas_rho(
        _val(state, constants, "P"),
        _val(state, constants, "R"),
        _val(state, constants, "T"),
    )


def _closure_boussinesq_rho(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("rho", "rho0", "beta", "dT")):
        return None
    return _val(state, constants, "rho") - Models.boussinesq_rho(
        _val(state, constants, "rho0"),
        _val(state, constants, "beta"),
        _val(state, constants, "dT"),
    )


def _closure_thermal_diffusivity(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("alpha", "k", "rho", "cp")):
        return None
    return _val(state, constants, "alpha") - Models.thermal_diffusivity(
        _val(state, constants, "k"),
        _val(state, constants, "rho"),
        _val(state, constants, "cp"),
    )


def _closure_kinematic_viscosity(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("nu", "mu", "rho")):
        return None
    return _val(state, constants, "nu") - Models.kinematic_viscosity(
        _val(state, constants, "mu"),
        _val(state, constants, "rho"),
    )


def _closure_mass_diffusivity(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("D", "fo_mass", "t", "L")):
        return None
    return _val(state, constants, "D") - Models.mass_diffusivity(
        _val(state, constants, "fo_mass"),
        _val(state, constants, "t"),
        _val(state, constants, "L"),
    )


def _closure_wave_speed_from_st(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("c", "omega", "L", "st_wave")):
        return None
    return _val(state, constants, "c") - Models.wave_speed_from_st(
        _val(state, constants, "omega"),
        _val(state, constants, "L"),
        _val(state, constants, "st_wave"),
    )


def _closure_dynamic_viscosity_from_re(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("mu", "rho", "u", "L", "re")):
        return None
    return _val(state, constants, "mu") - Models.dynamic_viscosity_from_re(
        _val(state, constants, "rho"),
        _val(state, constants, "u"),
        _val(state, constants, "L"),
        _val(state, constants, "re"),
    )


def _closure_scalar_diffusivity_from_pe(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("kappa", "u", "L", "pe")):
        return None
    return _val(state, constants, "kappa") - Models.scalar_diffusivity_from_pe(
        _val(state, constants, "u"),
        _val(state, constants, "L"),
        _val(state, constants, "pe"),
    )


def _closure_power_law_mu(state: Dict, constants: Dict) -> Optional[Any]:
    """mu_pred vs K * gamma_dot^(n-1). State: mu_pred (key mu_pl), gamma_dot, K, n."""
    if not _all_present(state, constants, ("mu_pl", "gamma_dot", "K", "n")):
        return None
    return _val(state, constants, "mu_pl") - Models.power_law_mu(
        _val(state, constants, "gamma_dot"),
        _val(state, constants, "K"),
        _val(state, constants, "n"),
    )


def _closure_arrhenius(state: Dict, constants: Dict) -> Optional[Any]:
    Rgas = _val(state, constants, "R_gas")
    if Rgas is None:
        Rgas = 8.314
    if not _all_present(state, constants, ("k_rate", "A", "Ea", "T")):
        return None
    return _val(state, constants, "k_rate") - Models.arrhenius_rate(
        _val(state, constants, "A"),
        _val(state, constants, "Ea"),
        _val(state, constants, "T"),
        R=Rgas,
    )


def _closure_stefan_boltzmann(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("q_rad", "epsilon", "T")):
        return None
    return _val(state, constants, "q_rad") - Models.stefan_boltzmann_flux(
        _val(state, constants, "epsilon"),
        _val(state, constants, "T"),
    )


def _closure_heat_flux_conduction(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("q_flux", "k", "dT", "dx")):
        return None
    return _val(state, constants, "q_flux") - Models.heat_flux_conduction(
        _val(state, constants, "k"),
        _val(state, constants, "dT"),
        _val(state, constants, "dx"),
    )


def _closure_speed_of_sound(state: Dict, constants: Dict) -> Optional[Any]:
    if not _all_present(state, constants, ("a", "gamma", "R", "T")):
        return None
    return _val(state, constants, "a") - Models.speed_of_sound(
        _val(state, constants, "gamma"),
        _val(state, constants, "R"),
        _val(state, constants, "T"),
    )


def _closure_specific_heat_nasa(state: Dict, constants: Dict) -> Optional[Any]:
    if _val(state, constants, "cp") is None or _val(state, constants, "T") is None:
        return None
    coeffs = _val(state, constants, "nasa_cp_coeffs")
    if coeffs is None:
        return None
    return _val(state, constants, "cp") - Models.specific_heat_nasa(
        _val(state, constants, "T"), coeffs
    )


# model_name -> list of (closure_id, fn)
CONSTITUTIVE_REGISTRY: Dict[str, List[Tuple[str, Callable[[Dict, Dict], Any]]]] = {
    "sutherland_mu": [("direct_mu", _closure_sutherland_direct)],
    "vft_mu": [("direct_mu", _closure_vft_direct)],
    "ideal_gas_rho": [("direct_rho", _closure_ideal_gas_rho)],
    "boussinesq_rho": [("direct_rho", _closure_boussinesq_rho)],
    "thermal_diffusivity": [("direct_alpha", _closure_thermal_diffusivity)],
    "kinematic_viscosity": [("direct_nu", _closure_kinematic_viscosity)],
    "mass_diffusivity": [("direct_D", _closure_mass_diffusivity)],
    "wave_speed_from_st": [("direct_c", _closure_wave_speed_from_st)],
    "dynamic_viscosity_from_re": [("direct_mu", _closure_dynamic_viscosity_from_re)],
    "scalar_diffusivity_from_pe": [("direct_kappa", _closure_scalar_diffusivity_from_pe)],
    "power_law_mu": [("direct_mu_pl", _closure_power_law_mu)],
    "arrhenius_rate": [("direct_k_rate", _closure_arrhenius)],
    "stefan_boltzmann_flux": [("direct_q_rad", _closure_stefan_boltzmann)],
    "heat_flux_conduction": [("direct_q_flux", _closure_heat_flux_conduction)],
    "speed_of_sound": [("direct_a", _closure_speed_of_sound)],
    "specific_heat_nasa": [("direct_cp", _closure_specific_heat_nasa)],
}


def list_constitutive_models() -> List[str]:
    """Names that can be passed to ResidualEngine(constitutive_audit=[...])."""
    return sorted(CONSTITUTIVE_REGISTRY.keys())


def run_constitutive_closures(
    model_names: List[str],
    state: Dict[str, Any],
    constants: Dict[str, Any],
    custom: Optional[List[Dict[str, Any]]] = None,
) -> Dict[str, Any]:
    """
    Returns nested dict flat under one level: {"model/closure_id": array, ...}.
    Skips closures that return None.
    """
    out: Dict[str, Any] = {}
    for name in model_names:
        for closure_id, fn in CONSTITUTIVE_REGISTRY.get(name, []):
            arr = fn(state, constants)
            if arr is not None:
                out[f"{name}/{closure_id}"] = jnp.asarray(arr)
    if custom:
        for spec in custom:
            cid = spec["name"]
            arr = spec["fn"](state, constants)
            if arr is not None:
                out[f"custom/{cid}"] = jnp.asarray(arr)
    return out



#!/usr/bin/env python3
"""
Generate a sample Physics Admissibility Report PDF for review.

Run from repo root after: pip install moju[report] or pip install reportlab
  python scripts/generate_sample_audit_pdf.py

Output: examples/sample_physics_admissibility_report.pdf
"""

import os
import sys

# Add repo root so moju.monitor.report is importable
_repo_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if _repo_root not in sys.path:
    sys.path.insert(0, _repo_root)

from moju.monitor.report import write_audit_pdf


def _sample_report() -> dict:
    """Build a sample report dict matching the structure returned by audit()."""
    # per_key r_norm = RMS(r)/scale with scale state-derived by default or overridden via r_ref
    return {
        "overall_admissibility_score": 0.92,
        "overall_admissibility_level": "High Admissibility",
        "per_key": {
            "laws/mass_incompressible": {
                "rms": 0.02,
                "r_norm": 0.08,
                "admissibility_score": 0.92,
                "admissibility_level": "High Admissibility",
            },
            "laws/navier_stokes": {
                "rms": 0.05,
                "r_norm": 0.14,
                "admissibility_score": 0.88,
                "admissibility_level": "Moderate Admissibility",
            },
            "scaling/re/chain_dx": {
                "rms": 0.01,
                "r_norm": 0.05,
                "admissibility_score": 0.95,
                "admissibility_level": "High Admissibility",
            },
            "scaling/pe/chain_dx": {
                "rms": 0.12,
                "r_norm": 0.43,
                "admissibility_score": 0.70,
                "admissibility_level": "Moderate Admissibility",
            },
            "scaling/pr/chain_dx": {
                "rms": 0.25,
                "r_norm": 0.83,
                "admissibility_score": 0.55,
                "admissibility_level": "Low Admissibility",
            },
            "constitutive/sutherland_mu/chain_dt": {
                "rms": 0.02,
                "r_norm": 0.09,
                "admissibility_score": 0.92,
                "admissibility_level": "High Admissibility",
            },
            "constitutive/ideal_gas_rho/ref_delta": {
                "rms": 0.01,
                "r_norm": 0.02,
                "admissibility_score": 0.98,
                "admissibility_level": "High Admissibility",
            },
            "data/u": {
                "rms": 0.08,
                "r_norm": 0.25,
                "admissibility_score": 0.80,
                "admissibility_level": "Moderate Admissibility",
            },
            "laws/heat_diffusion": {
                "rms": 0.45,
                "r_norm": 1.86,
                "admissibility_score": 0.35,
                "admissibility_level": "Non-Admissible",
            },
        },
    }


def main() -> None:
    examples_dir = os.path.join(_repo_root, "examples")
    os.makedirs(examples_dir, exist_ok=True)
    out_path = os.path.join(examples_dir, "sample_physics_admissibility_report.pdf")
    report = _sample_report()
    write_audit_pdf(
        report,
        out_path,
        model_name="NeuralTurb-Flowv3",
        model_id="223355592",
    )
    print(f"Sample PDF written to: {out_path}")


if __name__ == "__main__":
    main()



import jax
import jax.numpy as jnp

# Stefan-Boltzmann constant [W/(m^2 K^4)]; as JAX scalar for tracing
_STEFAN_BOLTZMANN = 5.670374419e-8

class Models:
    """
    Differentiable physical models for anchoring AI predictions to reality.
    Inputs may have leading batch dimensions; all operations broadcast and support
    both single-point and batched evaluation.
    """

    @staticmethod
    @jax.jit
    def sutherland_mu(T, mu0, T0, S):
        """
        Sutherland's Viscosity Law for Gases.
        
        :param T: Local temperature [K].
        :param mu0: Reference viscosity at T0 [Pa*s].
        :param T0: Reference temperature [K].
        :param S: Sutherland constant for the specific gas [K].
        :return: Dynamic viscosity at temperature T [Pa*s].
        """
        return mu0 * (T / T0) ** 1.5 * (T0 + S) / (T + S)

    @staticmethod
    @jax.jit
    def vft_mu(T, A, B, T0_v):
        """
        Vogel-Fulcher-Tammann (VFT) Viscosity Law for Liquids.
        
        :param T: Local temperature [K].
        :param A: Pre-exponential factor [Pa*s].
        :param B: Activation-related constant [K].
        :param T0_v: Vogel temperature (theoretical glass transition) [K].
        :return: Dynamic viscosity at temperature T [Pa*s].
        """
        return A * jnp.exp(B / (T - T0_v))

    @staticmethod
    @jax.jit
    def ideal_gas_rho(P, R, T):
        """
        Ideal Gas Law for Density.
        
        :param P: Absolute pressure [Pa].
        :param R: Specific gas constant [J/kg*K].
        :param T: Absolute temperature [K].
        :return: Density [kg/m^3].
        """
        return P / (R * T)

    @staticmethod
    @jax.jit
    def stefan_boltzmann_flux(epsilon, T):
        """
        Stefan-Boltzmann Radiative Heat Flux.
        
        :param epsilon: Surface emissivity (0 to 1).
        :param T: Absolute temperature [K].
        :return: Radiative heat flux [W/m^2].
        """
        return epsilon * _STEFAN_BOLTZMANN * T**4

    @staticmethod
    @jax.jit
    def boussinesq_rho(rho0, beta, dT):
        """
        Boussinesq Approximation for density variation.
        
        :param rho0: Reference density [kg/m^3].
        :param beta: Thermal expansion coefficient [1/K].
        :param dT: Temperature difference from reference [K].
        :return: Approximated density [kg/m^3].
        """
        return rho0 * (1 - beta * dT)

    @staticmethod
    @jax.jit
    def specific_heat_nasa(T, coeffs):
        """
        NASA 7-coefficient polynomial for specific heat (Cp).
        
        :param T: Temperature [K].
        :param coeffs: Array of 7 coefficients [a0, a1, a2, a3, a4, ...]; first 5 used for Cp/R polynomial.
        :return: Specific heat capacity [J/kg*K].
        """
        a = jnp.asarray(coeffs)
        return a[0] + a[1] * T + a[2] * T**2 + a[3] * T**3 + a[4] * T**4

    @staticmethod
    @jax.jit
    def power_law_mu(gamma_dot, K, n):
        """
        Non-Newtonian Power-Law viscosity.
        
        :param gamma_dot: Shear rate [1/s].
        :param K: Consistency index [Pa*s^n].
        :param n: Flow behavior index (dimensionless); n < 1 shear-thinning, n > 1 shear-thickening.
        :return: Apparent viscosity [Pa*s].
        
        Use case: Blood (n < 1), slurries (n > 1), shear-thinning fluids.
        """
        return K * gamma_dot ** (n - 1)

    @staticmethod
    @jax.jit
    def speed_of_sound(gamma, R, T):
        """
        Speed of sound in an ideal gas.
        
        :param gamma: Ratio of specific heats (dimensionless).
        :param R: Specific gas constant [J/kg*K].
        :param T: Absolute temperature [K].
        :return: Speed of sound [m/s].
        """
        return jnp.sqrt(gamma * R * T)

    @staticmethod
    @jax.jit
    def dynamic_pressure(rho, u):
        """
        Dynamic pressure (kinetic energy per unit volume).
        
        :param rho: Fluid density [kg/m^3].
        :param u: Flow velocity [m/s].
        :return: Dynamic pressure [Pa].
        """
        return 0.5 * rho * u**2

    @staticmethod
    @jax.jit
    def hydraulic_diameter(area, perimeter):
        """
        Hydraulic diameter for non-circular ducts.
        
        :param area: Cross-sectional flow area [m^2].
        :param perimeter: Wetted perimeter [m].
        :return: Hydraulic diameter [m].
        
        Use case: Reynolds number and friction in non-circular pipes.
        """
        return (4 * area) / perimeter

    @staticmethod
    @jax.jit
    def darcy_weisbach_dp(f, L, D, rho, u):
        """
        Darcy-Weisbach pressure drop in a pipe.
        
        :param f: Darcy friction factor (dimensionless).
        :param L: Pipe length [m].
        :param D: Pipe diameter [m].
        :param rho: Fluid density [kg/m^3].
        :param u: Mean flow velocity [m/s].
        :return: Pressure drop [Pa].
        """
        return f * (L / D) * (rho * u**2 / 2)

    @staticmethod
    @jax.jit
    def colebrook_friction(re, epsilon_d):
        """
        Haaland approximation for Darcy friction factor.
        
        :param re: Reynolds number (dimensionless).
        :param epsilon_d: Relative roughness (epsilon/D) (dimensionless).
        :return: Darcy friction factor (dimensionless).
        """
        return 1.0 / (-1.8 * jnp.log10((epsilon_d / 3.7) ** 1.11 + 6.9 / re)) ** 2

    @staticmethod
    @jax.jit
    def stokes_drag_force(mu, r, u):
        """
        Stokes drag force on a sphere (creeping flow, Re < 1).
        
        :param mu: Dynamic viscosity [Pa*s].
        :param r: Sphere radius [m].
        :param u: Relative flow velocity [m/s].
        :return: Drag force [N].
        """
        return 6 * jnp.pi * mu * r * u

    @staticmethod
    @jax.jit
    def kinematic_viscosity(mu, rho):
        """
        Kinematic viscosity (momentum diffusivity).
        
        :param mu: Dynamic viscosity [Pa*s].
        :param rho: Fluid density [kg/m^3].
        :return: Kinematic viscosity [m^2/s].
        """
        return mu / rho

    @staticmethod
    @jax.jit
    def thermal_diffusivity(k, rho, cp):
        """
        Thermal diffusivity of a material.
        
        :param k: Thermal conductivity [W/m*K].
        :param rho: Density [kg/m^3].
        :param cp: Specific heat capacity at constant pressure [J/kg*K].
        :return: Thermal diffusivity [m^2/s].
        """
        return k / (rho * cp)

    @staticmethod
    @jax.jit
    def mass_diffusivity(fo_mass, t, L):
        """
        Mass diffusivity recovered from Fourier number for mass.

        :param fo_mass: Mass Fourier number from Groups.fo_mass(D, t, L).
        :param t: Elapsed time [s].
        :param L: Characteristic length [m].
        :return: Mass diffusivity D [m^2/s].
        """
        return fo_mass * (L**2) / t

    @staticmethod
    @jax.jit
    def wave_speed_from_st(omega, L, st_wave):
        """
        Wave speed from wave Strouhal definition.

        :param omega: Angular frequency [rad/s].
        :param L: Characteristic length [m].
        :param st_wave: Wave Strouhal number from Groups.st_wave(omega, L, c).
        :return: Wave speed c [m/s].
        """
        return (omega * L) / st_wave

    @staticmethod
    @jax.jit
    def dynamic_viscosity_from_re(rho, u, L, re):
        """
        Dynamic viscosity from Reynolds number definition using local speed magnitude.

        :param rho: Density [kg/m^3].
        :param u: Velocity vector [m/s] or scalar speed.
        :param L: Characteristic length [m].
        :param re: Reynolds number.
        :return: Dynamic viscosity mu [Pa*s].
        """
        u_arr = jnp.asarray(u)
        u_mag = jnp.abs(u_arr) if u_arr.ndim == 0 else jnp.sqrt(jnp.sum(u_arr**2, axis=-1))
        return (jnp.asarray(rho) * u_mag * jnp.asarray(L)) / (jnp.asarray(re) + 1e-30)

    @staticmethod
    @jax.jit
    def scalar_diffusivity_from_pe(u, L, pe):
        """
        Effective scalar diffusivity from Peclet number definition.

        :param u: Velocity vector [m/s] or scalar speed.
        :param L: Characteristic length [m].
        :param pe: Peclet number.
        :return: Effective diffusivity kappa [m^2/s].
        """
        u_arr = jnp.asarray(u)
        u_mag = jnp.abs(u_arr) if u_arr.ndim == 0 else jnp.sqrt(jnp.sum(u_arr**2, axis=-1))
        return (u_mag * jnp.asarray(L)) / (jnp.asarray(pe) + 1e-30)

    @staticmethod
    @jax.jit
    def arrhenius_rate(A, Ea, T, R=8.314):
        """
        Arrhenius reaction rate constant vs temperature.
        
        :param A: Pre-exponential factor [1/s or appropriate units].
        :param Ea: Activation energy [J/mol].
        :param T: Absolute temperature [K].
        :param R: Universal gas constant [J/mol*K]; default 8.314.
        :return: Rate constant (same units as A).
        """
        return A * jnp.exp(-Ea / (R * T))

    @staticmethod
    @jax.jit
    def law_of_the_wall(y_plus):
        """
        Dimensionless velocity in the log-law region (turbulent boundary layer).
        
        :param y_plus: Dimensionless wall distance y_+ = y*u_tau/nu.
        :return: Dimensionless velocity u_+ = u/u_tau.
        
        Use case: Wall functions and boundary-layer modeling.
        """
        return 2.5 * jnp.log(y_plus) + 5.0

    @staticmethod
    @jax.jit
    def smagorinsky_nu_t(Cs, Delta, strain_rate_magnitude):
        """
        Smagorinsky subgrid eddy viscosity (scalar |S| form).

        :param Cs: Smagorinsky constant (dimensionless), typically ~0.1–0.2.
        :param Delta: Filter width [m].
        :param strain_rate_magnitude: Resolved strain-rate magnitude |S| [1/s]
            (often sqrt(2 S_ij S_ij) from the velocity field).
        :return: Eddy kinematic viscosity nu_t [m^2/s].

        Use case: LES-style closure auditing; combine with ``constitutive_audit`` chain
        residuals when ``strain_rate_magnitude`` varies in space or time.
        """
        return (Cs * Delta) ** 2 * strain_rate_magnitude

    @staticmethod
    @jax.jit
    def k_epsilon_nu_t(C_mu, k, epsilon, eps0):
        """
        Standard k–ε eddy viscosity (kinematic): νₜ = C_μ k² / (ε + ε₀).

        :param C_mu: Model constant (dimensionless), often ~0.09.
        :param k: Turbulent kinetic energy [m²/s²].
        :param epsilon: Turbulent dissipation rate [m²/s³].
        :param eps0: Positive floor on ε [same as epsilon] for numerical stability and
            stable JAX autodiff when ε → 0.
        :return: Eddy kinematic viscosity νₜ [m²/s].

        This closure is **algebraic νₜ only**. Full k–ε transport PDE residuals belong
        in ``Laws.*`` or a custom law, not in this helper.
        """
        return C_mu * k**2 / (epsilon + eps0)

    @staticmethod
    @jax.jit
    def k_omega_nu_t(k, omega, omega0):
        """
        Standard k–ω (Wilcox-style) eddy viscosity: νₜ = k / (ω + ω₀).

        :param k: Turbulent kinetic energy [m²/s²].
        :param omega: Specific dissipation rate [1/s].
        :param omega0: Positive floor on ω [same as omega] for stability and AD near ω → 0.
        :return: Eddy kinematic viscosity νₜ [m²/s].

        Algebraic νₜ only; ω and k transport equations are separate (``Laws.*`` / custom).
        """
        return k / (omega + omega0)

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_k_omega(u_laplacian, nu_molecular, k, omega, omega0):
        r"""Newtonian viscous acceleration :math:`(\nu_m+\nu_t)\nabla^2\mathbf{u}` with :math:`\nu_t=k/(\omega+\omega_0)`."""
        nu_t = Models.k_omega_nu_t(k, omega, omega0)
        nu_tot = jnp.asarray(nu_molecular) + nu_t
        ul = jnp.asarray(u_laplacian)
        if nu_tot.ndim < ul.ndim:
            nu_tot = nu_tot[..., jnp.newaxis]
        return nu_tot * ul

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_k_epsilon(u_laplacian, nu_molecular, C_mu, k, epsilon, eps0):
        r"""Same as :meth:`turbulent_viscous_acceleration_k_omega` but :math:`\nu_t=C_\mu k^2/(\varepsilon+\varepsilon_0)`."""
        nu_t = Models.k_epsilon_nu_t(C_mu, k, epsilon, eps0)
        nu_tot = jnp.asarray(nu_molecular) + nu_t
        ul = jnp.asarray(u_laplacian)
        if nu_tot.ndim < ul.ndim:
            nu_tot = nu_tot[..., jnp.newaxis]
        return nu_tot * ul

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_smagorinsky(u_laplacian, nu_molecular, Cs, Delta, strain_rate_magnitude):
        """Same with Smagorinsky :math:`\nu_t`."""
        nu_t = Models.smagorinsky_nu_t(Cs, Delta, strain_rate_magnitude)
        nu_tot = jnp.asarray(nu_molecular) + nu_t
        ul = jnp.asarray(u_laplacian)
        if nu_tot.ndim < ul.ndim:
            nu_tot = nu_tot[..., jnp.newaxis]
        return nu_tot * ul

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_compressible_k_omega(rho, u_laplacian, nu_molecular, k, omega, omega0):
        return jnp.asarray(rho)[..., jnp.newaxis] * Models.turbulent_viscous_acceleration_k_omega(
            u_laplacian, nu_molecular, k, omega, omega0
        )

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_compressible_k_epsilon(rho, u_laplacian, nu_molecular, C_mu, k, epsilon, eps0):
        return jnp.asarray(rho)[..., jnp.newaxis] * Models.turbulent_viscous_acceleration_k_epsilon(
            u_laplacian, nu_molecular, C_mu, k, epsilon, eps0
        )

    @staticmethod
    @jax.jit
    def turbulent_viscous_acceleration_compressible_smagorinsky(
        rho, u_laplacian, nu_molecular, Cs, Delta, strain_rate_magnitude
    ):
        return jnp.asarray(rho)[..., jnp.newaxis] * Models.turbulent_viscous_acceleration_smagorinsky(
            u_laplacian, nu_molecular, Cs, Delta, strain_rate_magnitude
        )

    @staticmethod
    @jax.jit
    def orifice_flow(Cd, A, dp, rho):
        """
        Volumetric flow rate through an orifice or restriction.
        
        :param Cd: Discharge coefficient (dimensionless).
        :param A: Orifice cross-sectional area [m^2].
        :param dp: Pressure drop across orifice [Pa].
        :param rho: Fluid density [kg/m^3].
        :return: Volumetric flow rate [m^3/s].
        """
        return Cd * A * jnp.sqrt(2 * dp / rho)

    @staticmethod
    @jax.jit
    def heat_flux_conduction(k, dT, dx):
        """
        Fourier's law: conductive heat flux.
        
        :param k: Thermal conductivity [W/m*K].
        :param dT: Temperature difference (positive for heat flow in +x) [K].
        :param dx: Distance over which dT is measured [m].
        :return: Heat flux in the x-direction [W/m^2].
        """
        return -k * (dT / dx)

    @staticmethod
    @jax.jit
    def surface_tension_eotvos(gamma0, T, Tc):
        """
        Eötvös rule: surface tension vs temperature (up to critical point).
        
        :param gamma0: Reference surface tension at T=0 [N/m].
        :param T: Temperature [K].
        :param Tc: Critical temperature [K].
        :return: Surface tension [N/m].
        
        Use case: Liquids; valid for T < Tc.
        """
        return gamma0 * (1 - T / Tc) ** (11 / 9)



#!/usr/bin/env python3
"""
End-to-end monitor example: 1D heat diffusion with NN T(t,x).

Shows:
  - Path A style state building (compute T, T_t, T_xx from the NN)
  - Laws.fourier_conduction residual + build_loss
  - Law-linked implied audit: ``thermal_diffusivity`` vs α implied from T_t / T_laplacian (automatic)
  - Optional chain audits on ``fo`` / ``bi`` (spatial/temporal consistency)
  - Required keys introspection
  - audit() PDF export when moju[report] installed
"""

import jax
import jax.numpy as jnp
import optax

from moju.piratio import Operators
from moju.piratio.groups import Groups
from moju.piratio.laws import Laws
from moju.monitor import ResidualEngine, audit, build_loss


def init_mlp(key, widths):
    params = []
    for m, n in zip(widths[:-1], widths[1:]):
        key, sub = jax.random.split(key)
        W = jax.random.normal(sub, (n, m)) * jnp.sqrt(2.0 / m)
        b = jnp.zeros((n,))
        params.append({"W": W, "b": b})
    return params


def mlp(params, tx):
    h = tx
    for layer in params[:-1]:
        h = jnp.tanh(h @ layer["W"].T + layer["b"])
    out = params[-1]
    return h @ out["W"].T + out["b"]


def T_field(params, t, x):
    # tx: (N,2) with columns [t,x]
    if t.ndim == 1 and x.ndim == 2:
        tx = jnp.concatenate([t[:, None], x], axis=-1)
    else:
        tx = jnp.concatenate([jnp.asarray(t)[..., None], jnp.asarray(x)], axis=-1)
    return mlp(params, tx)[..., 0]


def build_state(params, t, x, *, L, h, cp, k_solid, rho_ref):
    T = T_field(params, t, x)
    T_t = Operators.time_derivative(T_field, params, t, x)

    def lap_body(ti, xi):
        return Operators.laplacian(lambda p, x_in: T_field(p, ti, x_in), params, xi)

    T_xx = jax.vmap(lap_body)(t, x)

    def grad_body(ti, xi):
        g = Operators.gradient(lambda p, x_in: T_field(p, ti, x_in), params, xi)
        return g[0] if g.shape == (1,) else g

    T_x = jax.vmap(grad_body)(t, x)

    # Simple constitutive k(T), rho(T)
    kappa = k_solid * (1.0 + 0.001 * (T - 300.0))
    rho = rho_ref * (1.0 - 0.0001 * (T - 300.0))
    alpha = kappa / (rho * cp)

    # Derivatives needed for chain audits (same math as slab demo)
    dk_dT = 0.001 * k_solid
    drho_dT = -0.0001 * rho_ref
    d_k_dx = dk_dT * T_x
    d_k_dt = dk_dT * T_t
    d_rho_dx = drho_dT * T_x
    d_rho_dt = drho_dT * T_t
    d_alpha_dx = (1.0 / (rho * cp)) * d_k_dx - (kappa / (rho**2 * cp)) * d_rho_dx
    d_alpha_dt = (1.0 / (rho * cp)) * d_k_dt - (kappa / (rho**2 * cp)) * d_rho_dt

    Lb = jnp.broadcast_to(L, t.shape)
    hb = jnp.broadcast_to(h, t.shape)
    Fo = alpha * t / (Lb**2)
    Bi = hb * Lb / kappa
    d_t_dt = jnp.ones_like(t)
    d_Fo_dx = (t / (Lb**2)) * d_alpha_dx
    d_Fo_dt = (t / (Lb**2)) * d_alpha_dt + (alpha / (Lb**2)) * d_t_dt
    d_Bi_dx = -(hb * Lb / (kappa**2)) * d_k_dx
    d_Bi_dt = -(hb * Lb / (kappa**2)) * d_k_dt

    return {
        "T": T,
        "T_t": T_t,
        "T_xx": T_xx,
        "t": t,
        "L": Lb,
        "h": hb,
        "cp": jnp.broadcast_to(cp, t.shape),
        "kappa": kappa,
        "k": kappa,
        "rho": rho,
        "alpha": alpha,
        "k_solid": kappa,
        "Fo": Fo,
        "Bi": Bi,
        # derivatives
        "d_k_dx": d_k_dx,
        "d_k_dt": d_k_dt,
        "d_rho_dx": d_rho_dx,
        "d_rho_dt": d_rho_dt,
        "d_alpha_dx": d_alpha_dx,
        "d_alpha_dt": d_alpha_dt,
        "d_t_dt": d_t_dt,
        "d_Fo_dx": d_Fo_dx,
        "d_Fo_dt": d_Fo_dt,
        "d_Bi_dx": d_Bi_dx,
        "d_Bi_dt": d_Bi_dt,
    }


def main():
    L = 0.02
    h = 500.0
    cp = 900.0
    k_solid = 200.0
    rho_ref = 2700.0

    engine = ResidualEngine(
        constants={"cp": cp, "h": h},
        laws=[
            {
                "name": "fourier_conduction",
                "state_map": {"T_t": "T_t", "T_laplacian": "T_xx", "fo": "Fo", "t": "t", "L": "L"},
                "fn": Laws.fourier_conduction,
            }
        ],
        groups=[
            {"name": "fo", "state_map": {"alpha": "alpha", "t": "t", "L": "L"}, "output_key": "Fo", "fn": Groups.fo},
            {"name": "bi", "state_map": {"h": "h", "L": "L", "k_solid": "kappa"}, "output_key": "Bi", "fn": Groups.bi},
        ],
        # thermal_diffusivity implied_delta is prepended from fourier_conduction (law_implied_audits=True).
        scaling_audit=[
            {"name": "fo", "output_key": "Fo", "state_map": {"alpha": "alpha", "t": "t", "L": "L"}},
            {"name": "bi", "output_key": "Bi", "state_map": {"h": "h", "L": "L", "k_solid": "kappa"}},
        ],
    )

    print("Required state keys:", sorted(engine.required_state_keys()))

    key = jax.random.PRNGKey(0)
    params = init_mlp(key, [2, 32, 32, 1])
    opt = optax.adam(1e-3)
    opt_state = opt.init(params)

    # Collocation
    t = jnp.linspace(1.0, 10.0, 64)
    x = jnp.linspace(0.0, L, 64)[:, None]
    t_col, x_col = jnp.meshgrid(t, x[:, 0], indexing="ij")
    t_col = t_col.reshape(-1)
    x_col = x_col.reshape(-1, 1)

    def loss_fn(p):
        state = build_state(p, t_col, x_col, L=L, h=h, cp=cp, k_solid=k_solid, rho_ref=rho_ref)
        residuals = engine.compute_residuals(state, log_to_python=False)
        return build_loss(residuals)

    for _ in range(5):
        loss, grads = jax.value_and_grad(loss_fn)(params)
        updates, opt_state = opt.update(grads, opt_state)
        params = optax.apply_updates(params, updates)

    state = build_state(params, t_col, x_col, L=L, h=h, cp=cp, k_solid=k_solid, rho_ref=rho_ref)
    residuals = engine.compute_residuals(state, log_to_python=True)
    print("Residual categories:", residuals.keys())

    report = audit(engine.log, export_dir=".", last_residual_dict=residuals, save_residuals=True, model_name="Heat-1D", model_id="heat-demo")
    print("Overall score:", report["overall_admissibility_score"])
    print("Per-category:", report.get("per_category", {}))


if __name__ == "__main__":
    main()

